#*****************************************************************************
#
#    Sage Package "Modular Cohomology Rings of Finite Groups"
#
#    Copyright (C) 2009/2010
#         Simon A. King  <simon.king@nuigalway.ie> and
#         David J. Green <david.green@uni-jena.de>
#
#  Distributed under the terms of the GNU General Public License (GPL),
#  version 2 or later (at your choice)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************
r"""
Modular Cohomology Rings of Finite Groups

AUTHORS:

- Simon King  <simon.king@nuigalway.ie>
- David Green <david.green@uni-jena.de>

IMPLEMENTATION:

Our package is formed by Singular- and Gap-functions, as well as Python-, Cython- and C-code.

NOTE:

    Our package requires the `Small Groups
    <http://www-public.tu-bs.de:8080/~hubesche/small.html>`_ library
    of Hans Ulrich Besche, Bettina Eick and Eamonn O'Brien. If
    necessary, you can install it in Sage by
    ``install_package('database_gap')``.

The package "pGroupCohomology" comprises :func:`CohomologyRing`, a constructor
for the cohomology ring of a finite group `G` that does caching etc. behind the 
scenes. The package also comprises seven modules :mod:`~pGroupCohomology.cohomology`,
:mod:`~pGroupCohomology.modular_cohomology`, :mod:`~pGroupCohomology.barcode`,
:mod:`~pGroupCohomology.cochain`, :mod:`~pGroupCohomology.resolution`,
:mod:`~pGroupCohomology.mtx` and :mod:`~pGroupCohomology.dickson` that do the real work. 

In the first section of our introduction, we demonstrate the usage of
our package. After outlining the functionality, we go into more detail
and illustrate various ways of defining a cohomology ring, computing
the ring structure, and working in the cohomology ring.  Last, we
discuss how to save and load a cohomology ring on a local file system
and how to export and import it to a different platform.

In the second section, we give an overview on the theoretical
background of our package. The emphasis is on the construction of
parameters for the cohomology ring, which can then be used to prove
completeness of the ring structure by means of completeness criteria
that are due to different authors ([Benson]_, [GreenKing]_, [King]_,
[Symonds]_).  We illustrate this with step-by-step examples.

An overview of the various options that can be used is given in the
documentation of :func:`CohomologyRing`.

1. Synopsis
===========

Our package is devoted to the cohomology rings of finite groups with
coefficients in `\mathbb F_p` (the finite field with `p` elements,
where `p` is a prime less than 255 that divides the group order).

We expect the groups to be given either by their address in the Small
Groups Library, or as a group in Sage's Gap interface. In the latter
case and if `G` happens to be a prime power group, it is required that
the list of generators of `G` starts with a minimal generating set. If
this is not the case, a ``ValueError`` is raised. Internally, the group
will be transformed into a permutation group, which, in some cases, is
very difficult. So, it is recommended to start with a permutation group
right away.

The package is shipped with the cohomology rings for all groups of
order 64; the data are stored in folders and can simply be retrieved
by calling :func:`~pGroupCohomology.CohomologyRing` with appropriate
keys.  These data are provided to *all* users of the Sage installation,
thus we refer to it as the *public data base*.

In addition, we provide the cohomology rings for all groups of order
128 and for all but six groups of order 243 in the web. Unless the
user requires a recomputation from scratch, these data will be
automatically downloaded. In future versions, this will also be possible
for various non prime power groups.

Any user can construct new data bases; we refer to them as *private
data base*. There is a default location for a private data base,
namely in a sub-directory of ``DOT_SAGE``. But any location may be
chosen as well, provided that the user has write permissions.

In the prime power case, if the ring is not in the cache and unless
the user requests a different behaviour, the program looks for
pre-computed cohomology rings in the following default places, in
order:

1. In the public data base, which is in ``SAGE_DATA+'/pGroupCohomology/'``.
2. In the currently activated private data base.
3. In the currently activated cohomology internet source; if the
   data can be downloaded, they will be unpacked in the currently
   activated private data base.

If all this fails, a new ring is initialized. For groups of prime power orders,
this is first tried in the public data base. If it is impossible (e.g., since
the user does not have write permission), then the currently activated private
data base is used.

The data belonging to the cohomology ring of a prime power group (e.g., a minimal
free resolution, information on embedding of elementary abelian subgroups, standard
monomials etc.) are distributed in a folder tree, one tree for each group.

In the non prime power case, the situation is easier. It is not needed to
store small data pieces during the computation, and eventually the main
data are saved in a single file. Its default location is in the currently activated
private data base. The computation does, however, rely on the cohomology rings
of various subgroups, e.g., a Sylow subgroup. So, again the data are distributed.

Creating a cohomology ring
--------------------------

We describe here how to describe the group whose modular cohomology
ring shall be computed. There are various options to influence the
algorithm and the use of data bases; these options are described in
the documentation of :func:`~pGroupCohomology.CohomologyRing`.

... using the SmallGroups library
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

In our examples, we choose a temporary directory for our private data
base. Of course, in reality, a permanent directory would be chosen for
the data bases. We do so in order to avoid a conflict with an existing
private data base. For the same reason, we only address *existing*
entries in the public data base.
::

    sage: from pGroupCohomology import CohomologyRing
    sage: H0 = CohomologyRing(8,3)

Since the cohomology ring of ``SmallGroup(8,3)``, which is the
Dihedral Group of order 8, is in the public data base shipped with our
package, it is simply loaded and already completely computed. It is
not needed to specify a prime to determine the coefficients, since the
group is of prime power order. Note that we also have a list of group
names and group descriptions -- hence, ``H0`` knows that it belongs to
the Dihedral Group of order 8::

    sage: H0
    H^*(D8; GF(2))
    sage: print H0
    Cohomology ring of Dihedral group of order 8 with coefficients in GF(2)
    <BLANKLINE>
    Computation complete
    Minimal list of generators:
    [c_2_2: 2-Cocycle in H^*(D8; GF(2)),
     b_1_0: 1-Cocycle in H^*(D8; GF(2)),
     b_1_1: 1-Cocycle in H^*(D8; GF(2))]
    Minimal list of algebraic relations:
    [b_1_0*b_1_1]

The generators are of type :class:`~pGroupCohomology.cochain.COCH`.
It is also possible to convert the cohomology ring into a quotient
ring in the Singular interface::

    sage: singular(H0)
    //   characteristic : 2
    //   number of vars : 3
    //        block   1 : ordering M
    //                  : names    c_2_2 b_1_0 b_1_1
    //                  : weights      2     1     1
    //                  : weights     -1     0     0
    //                  : weights      0    -1     0
    //        block   2 : ordering C
    // quotient ring from ideal
    _[1]=b_1_0*b_1_1

In Sage, parent structures should be unique. And indeed, cohomology
rings of *the same* group in *the same* data base are unique::

    sage: H0 is CohomologyRing(8,3)
    True
    sage: H0 is loads(dumps(H0))
    True

We set up a private data base in a temporary directory, and show that
indeed two cohomology rings are only identical if their data are
stored in the same place::

    sage: tmp = tmp_filename()
    sage: CohomologyRing.set_user_db(tmp)
    sage: H1 = CohomologyRing.user_db(8,3, websource=False)
    sage: H1 is H0
    False

Since we used the option ``websource=False``, it was not attempted to
download the cohomology ring from the web. Hence, ``H1`` is not
completely computed yet, and therefore it is not even equal to
``H0``::

    sage: H0 == H1
    False

We now compute ``H1`` (see below for the theoretical background) and
verify that then ``H0`` is equal (though not identical) to ``H1``::

    sage: H1.make()
    sage: H0 == H1
    True

Here is a similar example, using a non prime power group. In this
case, we need to provide the field order by the optional parameter
``prime``. The generators are of type
:class:`~pGroupCohomology.cochain.MODCOCH`.

::

    sage: HS6a = CohomologyRing(720,763,prime=2)
    sage: HS6a.make()
    sage: print HS6a
    <BLANKLINE>
    Cohomology ring of SmallGroup(720,763) with coefficients in GF(2)
    <BLANKLINE>
    Computation complete
    Minimal list of generators:
    [c_2_1: 2-Cocycle in H^*(SmallGroup(720,763); GF(2)),
     c_1_0: 1-Cocycle in H^*(SmallGroup(720,763); GF(2)),
     b_3_3: 3-Cocycle in H^*(SmallGroup(720,763); GF(2)),
     c_3_2: 3-Cocycle in H^*(SmallGroup(720,763); GF(2))]
    Minimal list of algebraic relations:
    [b_3_3*c_3_2+c_2_1*c_1_0*b_3_3]
    <BLANKLINE>
    sage: singular(HS6a)
    //   characteristic : 2
    //   number of vars : 4
    //        block   1 : ordering M
    //                  : names    c_2_1 c_1_0 b_3_3 c_3_2
    //                  : weights      2     1     3     3
    //                  : weights     -1    -1     0    -1
    //                  : weights     -1     0     0     0
    //                  : weights      0    -1     0     0
    //        block   2 : ordering C
    // quotient ring from ideal
    _[1]=b_3_3*c_3_2+c_2_1*c_1_0*b_3_3

... using a group in the Gap interface
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Above, we mentioned that two cohomology rings are the same only if the
corresponding groups are the same. Indeed, if we represent the
Dihedral Group of order 8 in a different way then we obtain a
different ring. Note that, when referring to a group in the Gap
interface, we need to provide a name for that group, in one or the
other way. That name also determines the location of data storage, and
thus if two different names are provided then the corresponding
cohomology rings are not the same::

    sage: G = gap('DihedralGroup(8)')
    sage: H2 = CohomologyRing.user_db(G,GroupName='DihedralA')
    sage: G.SetName('"DihedralB"')
    sage: H3 = CohomologyRing.user_db(G)
    sage: H2 is H3
    False

Of course, when ``H2`` and ``H3`` are completely computed, they are
isomorphic to ``H1``, since they belong to isomorphic groups. However,
they do not have the same ring presentation. This is because the
presentation of the cohomology ring depends on the choice of a minimal
generating set of the group; and the minimal generating sets of
``gap('SmallGroup(8,3)')`` and of ``gap('DihedralGroup(8)')`` are
essentially different::

    sage: H2.make()
    sage: H3.make()
    sage: H2 == H3
    True
    sage: H2 == H1
    False

Indeed, the minimal relations of ``H1`` and ``H2`` look different::

    sage: print H1
    Cohomology ring of Dihedral group of order 8 with coefficients in GF(2)
    <BLANKLINE>
    Computation complete
    Minimal list of generators:
    [c_2_2: 2-Cocycle in H^*(D8; GF(2)),
     b_1_0: 1-Cocycle in H^*(D8; GF(2)),
     b_1_1: 1-Cocycle in H^*(D8; GF(2))]
    Minimal list of algebraic relations:
    [b_1_0*b_1_1]
    sage: print H2
    Cohomology ring of DihedralA with coefficients in GF(2)
    <BLANKLINE>
    Computation complete
    Minimal list of generators:
    [c_2_2: 2-Cocycle in H^*(DihedralA; GF(2)),
     b_1_0: 1-Cocycle in H^*(DihedralA; GF(2)),
     b_1_1: 1-Cocycle in H^*(DihedralA; GF(2))]
    Minimal list of algebraic relations:
    [b_1_1^2+b_1_0*b_1_1]

Above, we computed the cohomology of some non prime power group. It is
in fact a symmetric group::

    sage: G = gap('SymmetricGroup(6)')
    sage: HS6b = CohomologyRing(G, prime=2, GroupName='SymmetricGroup(6)')
    sage: HS6b.make()
    sage: print HS6b
    <BLANKLINE>
    Cohomology ring of SymmetricGroup(6) with coefficients in GF(2)
    <BLANKLINE>
    Computation complete
    Minimal list of generators:
    [c_2_1: 2-Cocycle in H^*(SymmetricGroup(6); GF(2)),
     c_1_0: 1-Cocycle in H^*(SymmetricGroup(6); GF(2)),
     b_3_2: 3-Cocycle in H^*(SymmetricGroup(6); GF(2)),
     b_3_3: 3-Cocycle in H^*(SymmetricGroup(6); GF(2))]
    Minimal list of algebraic relations:
    [b_3_2*b_3_3]
    <BLANKLINE>

Obviously the two ring presentations for the mod-2 cohomology of the
symmetric group of rank 6 are different. This is since the ring
presentation will essentially depend on the choice of generators of
the Sylow subgroup (which here is chosen by the program).


Computations in a cohomology ring
---------------------------------

Ring theoretic invariants
^^^^^^^^^^^^^^^^^^^^^^^^^

Up to now, we implemented the computation of four ring theoretic
invariants, namely the dimension, the depth, the a-invariants and
the Poincar\'e series::

    sage: H2.dimension()
    2
    sage: H2.depth()
    2
    sage: H2.a_invariants()
    [-Infinity, -Infinity, -2]
    sage: H2.poincare_series()
    1/(t^2 - 2*t + 1)

Of course, they are the same for both presentations of the cohomology
ring of the dihedral group of order 8::

    sage: H1.dimension()
    2
    sage: H1.depth()
    2
    sage: H1.a_invariants()
    [-Infinity, -Infinity, -2]
    sage: H1.poincare_series()
    1/(t^2 - 2*t + 1)

Actually we implemented two essentially different ways of computating
the Poincar\'e series. The method
:meth:`~pGroupCohomology.cohomology.COHO.poincare_without_parameters`
is usually much slower than
:meth:`~pGroupCohomology.cohomology.COHO.poincare_series`, but of
course the results are the same::

    sage: H2.poincare_without_parameters()
    1/(t^2 - 2*t + 1)

After computing the ring structure, generators and relations of the
cohomology ring are present. Generator number zero is the
multiplicative unit of the underlying field. The other generators are
represented by cochains (:class:`~pGroupCohomology.cochain.COCH`, or
:class:`~pGroupCohomology.cochain.MODCOCH` for non prime power groups)::

    sage: H1.gens()
    [1, c_2_2: 2-Cocycle in H^*(D8; GF(2)), b_1_0: 1-Cocycle in H^*(D8; GF(2)), b_1_1: 1-Cocycle in H^*(D8; GF(2))]

The relations are given as strings::

    sage: H1.rels()
    ['b_1_0*b_1_1']

The relations can be interpreted as elements of the cohomology ring,
and indeed they vanish::

    sage: print H1(H1.rels()[0])
    2-Cocycle in H^*(D8; GF(2)),
    represented by
    [0 0 0]

We can also compute the nil radical, i.e., the ideal formed by all
nilpotent elements. It is given in the Singular interface, as an ideal
in a free graded commututive ring. In particular, all elements of the
relation ideal of the cohomology ring are in the nil radical. For the
dihedral group of order 8, the nil radical is in fact equal to the
relation ideal::

    sage: H1.nil_radical()
    b_1_0*b_1_1
    sage: H2.nil_radical()
    b_1_1^2+b_1_0*b_1_1

Here are the data for the mod-2 cohomology of the symmetric group of
degree 6::

    sage: HS6a.dimension()
    3
    sage: HS6a.depth()
    3
    sage: HS6a.a_invariants()
    [-Infinity, -Infinity, -Infinity, -3]
    sage: HS6a.poincare_series()
    (-t^2 + t - 1)/(t^5 - 2*t^4 + t^3 - t^2 + 2*t - 1)
    sage: HS6a.poincare_without_parameters()
    (-t^2 + t - 1)/(t^5 - 2*t^4 + t^3 - t^2 + 2*t - 1)
    sage: HS6a.nil_radical()
    b_3_3*c_3_2+c_2_1*c_1_0*b_3_3


Induced maps
^^^^^^^^^^^^

As we have seen before, the presentation of the cohomology ring
depends on the choice of a minimal generating set of the underlying
finite `p`-group (resp. Sylow `p`-subgroup). For our purposes, we always
assume that a finite group has a fixed set of generators, and we call
two groups *equivalent*, if there is a group isomorphism that maps the
given list of generators of one group to an initial segment of the list
of generators of the other group.

:meth:`~pGroupCohomology.cohomology.COHO.group` returns a permutation
group in the gap interface equivalent to the one that was used to set up the
cohomology ring::

    sage: G1 = H1.group()
    sage: G1
    Group( [ (1,2)(3,8)(4,6)(5,7), (1,3)(2,5)(4,7)(6,8) ] )
    sage: G2 = H2.group()
    sage: G2
    Group( [ (1,5)(2,6)(3,8)(4,7), (1,3,2,4)(5,7,6,8) ] )

Apparently, ``G1`` and ``G2`` are different permutation groups. But in
fact they are isomorphic. Note that we could get a group isomorphism
by Gap's function ``IsomorphismGroups``. But the result is not
unique. In order to have reproducible doc tests, we provide an
explicit isomorphism::

    sage: phi = gap('GroupHomomorphismByImages( Group( [ (1,2)(3,8)(4,6)(5,7), (1,3)(2,5)(4,7)(6,8) ] ), Group( [ (1,5)(2,6)(3,8)(4,7), (1,3,2,4)(5,7,6,8) ] ), [ (1,2)(3,8)(4,6)(5,7), (1,3)(2,5)(4,7)(6,8) ], [ (1,5)(2,6)(3,8)(4,7), (1,8)(2,7)(3,6)(4,5) ] )')
    sage: phi.IsInjective()
    true
    sage: phi.IsSurjective()
    true

Of course, the group isomorphism induces an isomorphism of the
corresponding cohomology rings. After ensuring that ``phi`` is printed
nicely, we obtain the induced map::

    sage: phi.SetName('"phi"')
    sage: phi_star = H2.hom(phi,H1)
    sage: phi_star
    phi^*

Indeed, we can verify that the image of the relation of ``H2`` is zero
in ``H1``::

    sage: H2.rels()
    ['b_1_1^2+b_1_0*b_1_1']
    sage: phi_star(H2('b_1_1'))^2 + phi_star(H2('b_1_0'))*phi_star(H2('b_1_1'))
    (phi^*(b_1_1))**2+(phi^*(b_1_0))*(phi^*(b_1_1)): 2-Cocycle in H^*(D8; GF(2))
    sage: print phi_star(H2('b_1_1'))^2 + phi_star(H2('b_1_0'))*phi_star(H2('b_1_1'))
    2-Cocycle in H^*(D8; GF(2)),
    represented by
    [0 0 0]

We can also compute the inverse of ``phi_star``, compose the two maps,
and see that it is the identity map::

    sage: phi_star_inv = phi_star^(-1)
    sage: phi_star_inv
    (phi^*)^(-1)
    sage: [(phi_star_inv*phi_star)(X) == X for X in H2.gens()]
    [True, True, True, True]
    sage: [(phi_star*phi_star_inv)(X) == X for X in H1.gens()]
    [True, True, True, True]

It is possible to convert an induced homomorphism into a map of
quotient rings in the Singular interface. This allows for working with
cohomology ring elements of very high degrees. However, it is always
needed to take care of the ``basering`` in Singular::

    sage: S_phi_star = singular(phi_star)
    sage: singular(H2).set_ring()
    sage: I = singular.ideal(['c_2_2^50','b_1_0^50','b_1_1^50'])
    sage: singular(H1).set_ring()
    sage: S_phi_star(I)
    b_1_1^100+c_2_2^2*b_1_1^96+c_2_2^16*b_1_1^68+c_2_2^18*b_1_1^64+c_2_2^32*b_1_1^36+c_2_2^34*b_1_1^32+c_2_2^48*b_1_1^4+c_2_2^50,
    b_1_1^50+b_1_0^2*b_1_1^48+b_1_0^16*b_1_1^34+b_1_0^18*b_1_1^32+b_1_0^32*b_1_1^18+b_1_0^34*b_1_1^16+b_1_0^48*b_1_1^2+b_1_0^50,
    b_1_1^50
        
Note that Singular does not do automatic reduction in quotient
rings. So, eventually we do the reductions explicitly, in two ways,
with the same result::

    sage: S_phi_star(I).reduce(singular('ideal(0)'))
    b_1_1^100+c_2_2^2*b_1_1^96+c_2_2^16*b_1_1^68+c_2_2^18*b_1_1^64+c_2_2^32*b_1_1^36+c_2_2^34*b_1_1^32+c_2_2^48*b_1_1^4+c_2_2^50,
    b_1_1^50+b_1_0^50,
    b_1_1^50
    sage: singular(H2).set_ring()
    sage: I = I.reduce(singular('ideal(0)'))
    sage: singular(H1).set_ring()
    sage: S_phi_star(I)
    b_1_1^100+c_2_2^2*b_1_1^96+c_2_2^16*b_1_1^68+c_2_2^18*b_1_1^64+c_2_2^32*b_1_1^36+c_2_2^34*b_1_1^32+c_2_2^48*b_1_1^4+c_2_2^50,
    b_1_1^50+b_1_0^2*b_1_1^48+b_1_0^16*b_1_1^34+b_1_0^18*b_1_1^32+b_1_0^32*b_1_1^18+b_1_0^34*b_1_1^16+b_1_0^48*b_1_1^2+b_1_0^50,
    b_1_1^50+b_1_0*b_1_1^49+b_1_0^16*b_1_1^34+b_1_0^17*b_1_1^33+b_1_0^32*b_1_1^18+b_1_0^33*b_1_1^17+b_1_0^48*b_1_1^2+b_1_0^49*b_1_1
    sage: S_phi_star(I).reduce(singular('ideal(0)'))
    b_1_1^100+c_2_2^2*b_1_1^96+c_2_2^16*b_1_1^68+c_2_2^18*b_1_1^64+c_2_2^32*b_1_1^36+c_2_2^34*b_1_1^32+c_2_2^48*b_1_1^4+c_2_2^50,
    b_1_1^50+b_1_0^50,
    b_1_1^50

For a non prime power group, a permutation representation equivalent to the given group,
a subgroup which was used for the stable element method and a Sylow `p`-subgroup are available
as well::

    sage: HS6a.group()
    Group( [ (1,2), (1,2,3,4,5,6) ] )
    sage: HS6a.subgroup()
    Group( [ (1,2)(4,6), (1,6), (3,5), (1,6)(2,4) ] )
    sage: HS6a.sylow_subgroup()
    Group( [ (1,2)(4,6), (1,6), (3,5) ] )

The dihedral group of order 8 is a subgroup of the symmetric group of
rank 6. We compute the ring homomorphism that is induced by the
embedding. In order to get a reproducible result, we define the
group homomorphism explicitly::

    sage: phi = gap('GroupHomomorphismByImages( Group( [ (1,2)(3,8)(4,6)(5,7), (1,3)(2,5)(4,7)(6,8) ] ), Group( [ (1,2), (1,2,3,4,5,6) ] ), [ (1,2)(3,8)(4,6)(5,7), (1,3)(2,5)(4,7)(6,8) ], [ (1,3), (1,2)(3,4) ] )')
    sage: phi_star = HS6a.hom(phi,H0)
    sage: [H0.element_as_polynomial(phi_star(x)) for x in HS6a.gens()]   
    [1: 0-Cocycle in H^*(D8; GF(2)), b_1_1^2+c_2_2: 2-Cocycle in H^*(D8; GF(2)), b_1_0: 1-Cocycle in H^*(D8; GF(2)), c_2_2*b_1_1: 3-Cocycle in H^*(D8; GF(2)), 0: 3-Cocycle in H^*(D8; GF(2))]
    sage: singular(H0).set_ring()
    sage: singular(phi_star)
    s...[1]=b_1_1^2+c_2_2
    s...[2]=b_1_0
    s...[3]=c_2_2*b_1_1
    s...[4]=0


Massey products
^^^^^^^^^^^^^^^

We can compute (set valued) higher Massey products and the (single
valued) restricted Massey powers introduced by [Kraines]_ for cocycles
of prime power groups. For non prime power groups, we have so far only
an experimental implementation for restricted Massey powers.

We first study some examples on ``H0`` (cohomology of the Dihedral Group
of order 8) that we have defined above. Note that we are transforming
the returned sets into sorted lists, in order
to have reproducible doc tests::

    sage: sorted(list(H0.massey_products(H0.2,H0.3,H0.2)), cmp=lambda X,Y:cmp(X.name(),Y.name()))
    [0: 2-Cocycle in H^*(D8; GF(2)), b_1_0^2: 2-Cocycle in H^*(D8; GF(2))]
    sage: sorted(list(H0.massey_products(H0.2,H0.3,H0.2,H0.3)), cmp=lambda X,Y:cmp(X.name(),Y.name()))
    [b_1_0^2+c_2_2: 2-Cocycle in H^*(D8; GF(2)),
     b_1_1^2+b_1_0^2+c_2_2: 2-Cocycle in H^*(D8; GF(2)),
     b_1_1^2+c_2_2: 2-Cocycle in H^*(D8; GF(2)),
     c_2_2: 2-Cocycle in H^*(D8; GF(2))]
    sage: sorted(list(H0.massey_products(H0.2*H0.1,H0.3*H0.1,H0.2*H0.1)), cmp=lambda X,Y:cmp(X.name(),Y.name())) # long time
    [0: 8-Cocycle in H^*(D8; GF(2)),
     c_2_2*b_1_0^6: 8-Cocycle in H^*(D8; GF(2)),
     c_2_2*b_1_0^6+c_2_2^2*b_1_0^4: 8-Cocycle in H^*(D8; GF(2)),
     c_2_2*b_1_0^6+c_2_2^2*b_1_0^4+c_2_2^3*b_1_0^2: 8-Cocycle in H^*(D8; GF(2)),
     c_2_2*b_1_0^6+c_2_2^3*b_1_0^2: 8-Cocycle in H^*(D8; GF(2)),
     c_2_2^2*b_1_0^4: 8-Cocycle in H^*(D8; GF(2)),
     c_2_2^2*b_1_0^4+c_2_2^3*b_1_0^2: 8-Cocycle in H^*(D8; GF(2)),
     c_2_2^3*b_1_0^2: 8-Cocycle in H^*(D8; GF(2))]

If one is interested in just one element of the Massey products, one
can use the option ``all=False``::

    sage: H0.massey_products(H0.2*H0.1,H0.3*H0.1,H0.2*H0.1,H0.3*H0.1, all=False)
    set([c_2_2^5: 10-Cocycle in H^*(D8; GF(2))])

By the `i`-th restricted Massey power of a cocycle `C` in a cohomology
ring with coefficients `\mathbb F_p`, we mean the restricted
`p^i`-fold Massey product of `C` as introduced by D. Kraines. If it is
defined, then it is a single cocycle::

    sage: H = CohomologyRing.user_db(27,3)
    sage: H.make()
    sage: H.8
    a_3_4: 3-Cocycle in H^*(E27; GF(3))
    sage: H.8.massey_power()
    <a_3_4; 1>: 8-Cocycle in H^*(E27; GF(3))
    sage: H.element_as_polynomial(_)
    b_2_0^2*a_1_1*a_3_5+b_2_0^2*a_1_0*a_3_4+b_2_0*c_6_8: 8-Cocycle in H^*(E27; GF(3))

It is known that in odd degree and for `p>2` it can be expressed as
minus the composition of the first Steenrod power with the Bockstein
operator. This allows for a verification of the above result, see
:meth:`~pGroupCohomology.cochain.COCH.massey_power` for details.

Massey products allow to distinguish certain isomorphic cohomology
rings, so, they contain information that go beyond the ring
structure. An easy example is provided by the cohomology rings of the
cyclic groups of order 3 and order 9, which are isomorphic.
::

    sage: HC3 = CohomologyRing.user_db(3,1)
    sage: HC3.make()
    sage: HC9 = CohomologyRing.user_db(9,1)
    sage: HC9.make()
    sage: print HC3
    Cohomology ring of Small Group number 1 of order 3 with coefficients in GF(3)
    <BLANKLINE>
    Computation complete
    Minimal list of generators:
    [c_2_0: 2-Cocycle in H^*(SmallGroup(3,1); GF(3)),
     a_1_0: 1-Cocycle in H^*(SmallGroup(3,1); GF(3))]
    Minimal list of algebraic relations:
    []
    sage: print HC9
    Cohomology ring of Small Group number 1 of order 9 with coefficients in GF(3)
    <BLANKLINE>
    Computation complete
    Minimal list of generators:
    [c_2_0: 2-Cocycle in H^*(SmallGroup(9,1); GF(3)),
     a_1_0: 1-Cocycle in H^*(SmallGroup(9,1); GF(3))]
    Minimal list of algebraic relations:
    []

Note that the relation ``a_1_0^2`` is not listed, since the relation
`x^2=0` holds for any cocycle `x` of odd degree, for `p>2`, and is
thus somehow redundant. The 1st restricted Massey power of the degree
one generator is non-trivial for the cyclic group of order 3, but only
the 2nd restricted Massey power of the degree one generator of the
cohomology ring of the cyclic group of order 9 is non-trivial::

    sage: HC3.element_as_polynomial(HC3.2.massey_power())
    -c_2_0: 2-Cocycle in H^*(SmallGroup(3,1); GF(3))
    sage: HC9.element_as_polynomial(HC9.2.massey_power())
    0: 2-Cocycle in H^*(SmallGroup(9,1); GF(3))
    sage: HC9.element_as_polynomial(HC9.2.massey_power(2))
    -c_2_0: 2-Cocycle in H^*(SmallGroup(9,1); GF(3))

Bar Codes
^^^^^^^^^

Cohomology rings contain much information on a group. Unfortunately,
it is, in general, quite difficult to test whether two graded
commutative rings are isomorphic. Hence, it is of interest to consider
invariants that can be more easily compared, such as the Poincar\'e
series or the depth. Bar codes are not ring theoretic invariants,
but are defined in terms of the finite group whose cohomology is
being studied.

Any sequence of group homomorphisms `\phi_i: G_i \to G_{i+1}` with
`i=1,...,n` gives rise to a series of induced homomorphisms of
cohomology rings. Persistent group cohomology, introduced by Graham
Ellis <graham.ellis@nuigalway.ie> and Simon King
<simon.king@nuigalway.ie>, asks how long cocycles in a given degree
"survive" being mapped by the induced homomorphisms. A preprint is
in preparation.

Note that we insist that the sequence starts with the group and ends
with the trivial group. It is easy to get such sequences in the case
of prime power groups (e.g., upper central series, derived series,
`p`-central series). For non prime power groups, this may be more
difficult.

For a given degree `d`, the persistent cohomology can be described by
an upper triangular matrix of non-negative integers, where entry `i,j`
(`i\le j`) is the dimension of the image of the degree `d` part of
`H^*(G_j)` in `H^*(G_i)` under the induced homomorphism given by the
composition of `\phi_i,\phi_{i+1},...,\phi_{j-1}`, including the
case `i=j` in which the matrix entry simply gives the dimension of the
degree `d` part of `H^*(G_i)`.

As usual, the sequence of dimensions sorted by degree gives rise to a
Poincar\'e series. Hence, an upper triangular matrix of rational
functions results and this provides information for all degrees.

We like to emphasize that the persistence matrices are very easy to
compare and provide a quite strong tool for distinguishing groups.
While the cohomology ring only depends on the group algebra, the bar
codes may contain additional information on the group itself, by
taking into account the normal series. However, the following example
shows that bar codes contain information that even goes beyond the
isomorphism types of the normal subgroups and quotient groups.
::

    sage: H158 = CohomologyRing(64,158)
    sage: H160 = CohomologyRing(64,160)

These cohomology rings are part of our data base, hence, they are
already completely known. The Poincar\'e series, the a-invariants, the
degrees of generators and of relations of the cohomology rings
coincide::

    sage: H158.poincare_series()
    (t^4 + t^3 + t^2 + t + 1)/(t^6 - 2*t^5 + 3*t^4 - 4*t^3 + 3*t^2 - 2*t + 1)
    sage: H160.poincare_series()
    (t^4 + t^3 + t^2 + t + 1)/(t^6 - 2*t^5 + 3*t^4 - 4*t^3 + 3*t^2 - 2*t + 1)
    sage: H158.degvec
    [4, 4, 1, 1, 1, 3, 3]
    sage: H160.degvec
    [4, 4, 1, 1, 1, 3, 3]
    sage: H158.set_ring()
    sage: [singular('deg(%s)'%r) for r in H158.rels()]
    [2, 2, 3, 3, 4, 4, 5, 6, 6, 6]
    sage: H160.set_ring()
    sage: [singular('deg(%s)'%r) for r in H160.rels()]
    [2, 2, 3, 3, 4, 4, 5, 6, 6, 6]
    sage: H158.a_invariants()
    [-Infinity, -Infinity, -2]
    sage: H160.a_invariants()
    [-Infinity, -Infinity, -2]

We consider here the bar codes associated with the upper central
series. It turns out that the non-trivial terms of the upper central
series and the resulting factor groups are isomorphic::

    sage: G158 = gap('SmallGroup(64,158)')
    sage: G160 = gap('SmallGroup(64,160)')
    sage: [(G.IdGroup(), (G158/G).IdGroup()) for G in G158.UpperCentralSeries()]
    [([ 64, 158 ], [ 1, 1 ]),
     ([ 16, 2 ], [ 4, 2 ]),
     ([ 4, 2 ], [ 16, 11 ]),
     ([ 1, 1 ], [ 64, 158 ])]
    sage: [(G.IdGroup(), (G160/G).IdGroup()) for G in G160.UpperCentralSeries()]
    [([ 64, 160 ], [ 1, 1 ]),
     ([ 16, 2 ], [ 4, 2 ]),
     ([ 4, 2 ], [ 16, 11 ]),
     ([ 1, 1 ], [ 64, 160 ])]

Nevertheless, the groups can be distinguished using the bar codes
associated with the upper central series::

    sage: B158 = H158.bar_code('UpperCentralSeries')
    sage: B160 = H160.bar_code('UpperCentralSeries')
    sage: B158 == B160
    False

Indeed, the bar codes differ in degree 3; graphically::

    sage: print B158[3]
            *
            *
          *-*
          *-*
          *
          *
          *
          *
          *
          *
        *-*
        *-*
        *
        *
      *
      *
      *
      *
    *
    *
    *
    *
    sage: print B160[3]
            *
            *
          *-*
          *
          *
          *
          *
          *
          *
          *
        *-*-*
        *-*
        *
      *-*
      *
      *
      *
    *
    *
    *
    *

These pictures (bar codes) can be interpreted as follows. Let `G` be a
finite `p`-group and let `G=G_0 > G_1 > ... > G_n > 1` be a normal
series; in our example, we have `n=2`. The first `n+1` columns of the
bar code correspond to the normal subgroups groups `G_n, G_{n-1},...,
G_0`, while the last `n` columns correspond to the factor groups
`G/G_n, G/G_{n-1},..., G/G_1`. We consider the sequence of group
homomorphisms given by inclusions and quotients. The stars in the bar
code correspond to basis vectors of the degree `d` part of the
cohomology rings of the respective groups. Stars are connected by a
line (i.e., a hyphen) if the corresponding basis vectors are mapped to
each other by the induced maps (which of course go from right to
left).

In terms of the persistence matrix::

    sage: B158[3].matrix()
    [ 4  0  0  0  0]
    [ 0  4  0  0  0]
    [ 0  0  4  2  0]
    [ 0  0  0 10  2]
    [ 0  0  0  0  4]
    sage: B160[3].matrix()
    [ 4  0  0  0  0]
    [ 0  4  1  0  0]
    [ 0  0  4  2  1]
    [ 0  0  0 10  2]
    [ 0  0  0  0  4]

The bar codes in degree 3 can also be computed directly, and of course
the result coincides with the bar codes obtained from the"full" bar codes 
(based on Poincar\'e series)::

    sage: B158[3] == H158.bar_code('UpperCentralSeries',degree=3)
    True
    sage: B160[3] == H160.bar_code('UpperCentralSeries',degree=3)
    True

Since apparently the ``(2,4)`` entries of the persistence matrices
differ, we show here how the Poincar\'e series in that position look
like::

    sage: B158.matrix()[2,4]
    2*t^2 + 2*t + 1
    sage: B160.matrix()[2,4]
    t^3 + 2*t^2 + 2*t + 1

Hence, the iamge of the map `H^*(G/G_2) \to H^*(G)` induced by the
quotient map contains only finitely many cocycles, where `G` is group
number 158 or 160 of order 64, and `G_2` is the second term of the
upper central series.


Data storage and data relocation
--------------------------------

Modular cohomology rings of finite groups can be very
complicated. When computing the ring structure, many data pieces are
needed --- but not all pieces at the same time. Therefore we save the
various pieces in separate files, and reload them when they are
needed. In some examples, the memory and time consumption of saving
*all* data in a single file would not be reasonable. Hence, 'saving' a
cohomology ring to some file ``f`` means that ``f`` only contains the
information needed to find the files in which the various data pieces
are stored.

This method is fast, but has a clear disadvantage: If you wish to
transfer your cohomology computations to a different computer, it does
not suffice to copy one file for one cohomology ring. *All* data
pieces must be copied. Moreover, when the data are in a different
location, then the information stored in ``f`` would point to
non-existing paths. We tried to address these problems in the least
inconvenient way, as explained in the following sub-sections.

Data storage
^^^^^^^^^^^^

We first describe the situation for prime power groups. We do some new
computations from scratch, using a temporary directory.
::
  
    sage: from pGroupCohomology import CohomologyRing
    sage: tmp = tmp_filename()
    sage: CohomologyRing.set_user_db(tmp)
    sage: H = CohomologyRing.user_db(8,3,websource=False)
    sage: H.make()

By default, when computing a cohomology ring ``H``, the result is
automatically saved in a file ``H.autosave_name()``. Cohomology rings
are unique parent structures, and therefore reloading from that file
only creates a new reference to ``H``::

    sage: K = load(H.autosave_name())
    sage: K is H
    True
    sage: loads(dumps(H)) is H
    True

By the command ``CohomologyRing.set_user_db(tmp)``, all data in the
user data base are stored in sub-folders of the folder ``tmp``. The
names of these folders are composed as follows.

 1. The basic folder ``tmp`` is known to ``H`` by the attribute
    ``root``. :func:`~pGroupCohomology.CohomologyRing` normalizes
    the given folder name ``tmp`` so that it does not contain double
    slashes. In order to not break the following doc test, we remove
    double slashes from ``tmp``.
    ::

        sage: tmp = tmp.replace('//','/')
        sage: H.root == tmp+'/'
        True

 2. Any cohomology ring has been assigned a "stem name" that determines
    the names of sub-folders and file names associated with the ring.
    If the cohomology ring belongs to a group in the Small Groups Library,
    its stem name is formed using the address in the Small Groups Library.
    Otherwise, it is obtained from the group name used in the definition
    of the cohomology ring::

        sage: H = CohomologyRing.user_db(8,3)
        sage: G = gap('DihedralGroup(8)')
        sage: K = CohomologyRing.user_db(G, GroupName = 'DihedralA')
        sage: G.SetName('"DihedralB"')
        sage: L = CohomologyRing.user_db(G)
        sage: H.GStem
        '8gp3'
        sage: K.GStem
        'DihedralA'
        sage: L.GStem
        'DihedralB'

 3. Data belonging to a cohomology ring can be found in a folder that is 
    stored as the attribute "gps_folder", and is composed by the root and
    the stem name::

        sage: H.gps_folder == H.root+H.GStem+'/'
        True

The situation for non prime power groups is different. Here, the data
structure for the ring itself is somewhat easier, so, main data are
saved in just one file.  On the other hand, the computations rely on
the stable element method and thus involve the cohomology rings of
various subgroups. When reloading, these are searched in the cache or
in the public or private data bases or in the web repository, and are
re-computed if necessary. Here, we show that the rings are cached::

    sage: HS6a = CohomologyRing(720,763,prime=2)
    sage: HS6a.make()
    sage: HS6a is loads(dumps(HS6a))
    True
    sage: save(HS6a, HS6a.autosave_name())
    sage: HS6a is load(HS6a.autosave_name())
    True

Data relocation
^^^^^^^^^^^^^^^

If you wish to transfer data to a different location, you may do
so. But please make sure that (in the case of prime power groups)
``H.gps_folder`` keeps its name (only the parent directory is allowed
to change!), and that all files in ``H.gps_folder`` and its
sub-folders are preserved. Then, by ``CohomologyRing.set_user_db``,
one can activate the private data base in the new location, and reload
the ring.  In order to demonstrate that it really is the same ring, we
provide ``H`` with an additional attribute before saving and moving
the data::

    sage: H.setprop('info', 'The cohomology data were moved')
    sage: save(H,H.autosave_name())
    sage: tmp2 = tmp_filename()
    sage: import os
    sage: os.rename(tmp,tmp2)
    sage: CohomologyRing.set_user_db(tmp2)
    sage: M = CohomologyRing.user_db(8,3)
    sage: print M
    Cohomology ring of Dihedral group of order 8 with coefficients in GF(2)
    <BLANKLINE>
    Computation complete
    Minimal list of generators:
    [c_2_2: 2-Cocycle in H^*(D8; GF(2)),
     b_1_0: 1-Cocycle in H^*(D8; GF(2)),
     b_1_1: 1-Cocycle in H^*(D8; GF(2))]
    Minimal list of algebraic relations:
    [b_1_0*b_1_1]
    sage: M.info
    'The cohomology data were moved'

2. Background for the Cohomology Computation of Prime Power Groups
==================================================================

Elementary abelian subgroups and choice of generators
-----------------------------------------------------

Let `G` be a non-abelian finite `p`-group that is not a generalised
quaternion group. Let `S_1` be its greatest central elementary abelian
subgroup, and let `S_2,...,S_n` be representatives for the conjugacy
classes of maximal elementary abelian subgroups of `G`.

Our approach relies on knowing these subgroups, to which we refer as
*special subgroups*. When defining a cohomology ring, the special
subgroups are computed using Gap functions written by David Green.

In our example, `G` is the dihedral group of order 8, which is the
third group of order 8 in the Small Groups Library. We set up the
cohomology ring in a temporary directory. The special subgroups can be
retrieved as follows::

    sage: from pGroupCohomology import CohomologyRing
    sage: tmp = tmp_filename()
    sage: CohomologyRing.set_user_db(tmp)
    sage: H = CohomologyRing.user_db(8,3, websource=False)
    sage: sorted(H.subgroups().items())
    [((2, 1), H^*(SmallGroup(2,1); GF(2))), ((4, 2), H^*(SmallGroup(4,2); GF(2)))]

Hence, ``H.subgroups()`` is a dictionary, whose keys correspond to
elementary abelian groups in the SmallGroups library, and whose values
are the corresponding cohomology rings.

[Green]_ provides an algorithm for the computation of minimal
projective resolutions of finite `p`-groups, using non-commutative
Groebner bases. He implemented it in C-programs, and we provide a
Cython wrapper in :mod:`pGroupCohomology.resolution`.

Knowing a resolution out to degree `d` allows for computing a minimal
set of cohomology generators out to degree `d` and a minimal set of
algebraic relations of degree up to `d` for these generators. It pays
off to choose the generators in a special way, as follows:

1. A cohomology class of `G` is nilpotent if and only if its 
   restrictions to `S_2,...,S_n` are nilpotent. Since the cohomology
   rings of elementary abelian groups are either polynomial rings
   or the tensor product of a polynomial ring with an outer algebra
   (if `p` is odd), it is easy to test whether a cohomology class of `G`
   is nilpotent, even if the whole cohomology ring is not known yet.
   We choose as many as possible new nilpotent generators in degree `d`.
2. After step 1, continue with choosing as many as possible new
   generators in degree `d` with nilpotent restriction to `S_1`.
3. After step 2, include the remaining new generators in degree 
   `d`. These are Duflot regular.

If `z` is the rank of the centre of `G` then there are exactly `z`
Duflot regular generators, which was proved by [Kuhn]_.  David Green
introduced a monomial ordering on the cohomology ring of `G` that, in
the first place, depends on the degree of a monomial, but also depends
on the number of nilpotent respectively Duflot regular generators
occuring in the monomial. It turns out that this monomial ordering
magically simplifies many computations. We now compute the cohomology
ring ``H`` out to degree 2, and show the generators and relations.
::

    sage: H.make(2)
    sage: H.gens()
    [1,
     c_2_2: 2-Cocycle in H^*(D8; GF(2)),
     b_1_0: 1-Cocycle in H^*(D8; GF(2)),
     b_1_1: 1-Cocycle in H^*(D8; GF(2))]
    sage: H.rels()
    ['b_1_0*b_1_1']

We look at the generators in more detail. Generator number ``i``
(``i`` an integer) of ``H`` can be obtained by ``H.gen(i)`` or by
``H.i`` (the latter only if ``i`` is a *number*, not just a *variable*
assigned to a number). Note that ``H.0`` always returns the
multiplicative unit in the underlying base field, which is ``GF(2)``
in this case.
::

    sage: H.0
    1
    sage: print H.1
    2-Cocycle in H^*(D8; GF(2)),
    represented by
    [0 0 1]
    rdeg = 1
    ydeg = 0
    sage: print H.2
    1-Cocycle in H^*(D8; GF(2)),
    represented by
    [1 0]
    rdeg = 0
    ydeg = 0
    sage: print H.3
    1-Cocycle in H^*(D8; GF(2)),
    represented by
    [0 1]
    rdeg = 0
    ydeg = 0

``rdeg`` and ``ydeg`` are notions needed in the definition of David
Green's monomial ordering. ``H.1.rdeg()==1`` means that the generator
``H.1`` is Duflot regular. We now compute the restrictions. The
restriction maps can be obtained as follows::

    sage: sorted(H.restriction_maps().items())
    [(1,
      [(2, 1),
       Induced homomorphism of degree 0 from H^*(D8; GF(2)) to H^*(SmallGroup(2,1); GF(2))]),
     (2,
      [(4, 2),
       Induced homomorphism of degree 0 from H^*(D8; GF(2)) to H^*(SmallGroup(4,2); GF(2))]),
     (3,
      [(4, 2),
       Induced homomorphism of degree 0 from H^*(D8; GF(2)) to H^*(SmallGroup(4,2); GF(2))])]

Hence, it is a dictionary (in order to have a reproducible doc test,
we needed to transform it into a sorted list). The keys correspond to
the number of a special subgroup and yield a pair formed by the
SmallGroups identifier of that subgroup and the restriction map. It
can be applied to a generator in multiplicative or functional
notation::

    sage: print H.restriction_maps()[1][1]*H.1
    2-Cocycle in H^*(SmallGroup(2,1); GF(2)),
    represented by
    [1]
    sage: print H.restriction_maps()[1][1](H.1)
    2-Cocycle in H^*(SmallGroup(2,1); GF(2)),
    represented by
    [1]

As required, the restriction of the Duflot regular generator ``H.1``
to `S_1` is not nilpotent. Actually, for `p=2`, the restriction of a
cohomology class to some elementary abelian subgroup is nilpotent if
and only if the restriction is zero.

The list of generators restricted to any subgroup and expressed as a
polynomial in the generators of the subgroup's cohomology ring can be
obtained by
:meth:`~pGroupCohomology.cohomology.COHO.restrict_to_subgroup`::

    sage: H.restrict_to_subgroup(2)
    [c_1_0*c_1_1+c_1_0^2: 2-Cocycle in H^*(SmallGroup(4,2); GF(2)),
     c_1_1: 1-Cocycle in H^*(SmallGroup(4,2); GF(2)),
     0: 1-Cocycle in H^*(SmallGroup(4,2); GF(2))]

Note that the choice of generators and relations depends on how the
group was originally defined, as we have seen above.

Dave Benson's completeness criterion
------------------------------------

Dickson classes
^^^^^^^^^^^^^^^

One key problem is to detect whether all generators and relations are
already known. In the case of an abelian group, it is known that a
computation out to degree 2 will always suffice.  In the case of
generalized quaternion groups, it is known that a computation out to
degree 4 will always suffice. For general prime power groups, we solve
the termination problem according to a method introduced by [Benson]_,
improved by [GreenKing]_, and by another criterion introduced by
[Symonds]_. Note that we use a third criterion in the non prime power
case, as described below.

Let `r` be the `p`-rank of `G`, and let `z` be the rank of the centre
of `G`. Using Dickson invariants, we define `r-z` cohomology classes
`d_{1,j},...,d_{r-z,j}` of `S_j` (`j=2,...,n`) in the polynomial part
of the cohomology ring of `S_j`, restricting to zero on `S_1` (recall
that `S_2,...S_n` are maximal elementary abelian subgroups of `G`, and
`S_1` is a common subgroup, namely the greatest central elementary
abelian subgoup). 

One can show that there exist cohomology classes `D_1,...,D_{r-z}` of
`G` such that `D_i` simultaneously restricts to `d_{i,2},...,d_{i,n}`
or to their `p^{k_i}`-th power for some `k_1,...,k_{r-z}`.  We briefly
refer to these classes as *restricted Dickson classes* or
*restricted Dickson elements* for `G` respectively for `S_j` (or we even
drop the word *restricted*, if it is clear from the context).

Using results of [Benson]_, [GreenKing]_ show that the Duflot regular
generators of `H^*(G)` together with the restricted Dickson classes
for `G` form a filter-regular homogeneous system of parameters (HSOP)
for `H^*(G)`. Filter-regular means that the multiplication map by
parameter number `k` on the quotient of `H^*(G)` modulo the first `k-1`
parameters has finite dimensional kernel. Since they are parameters,
the quotient of `H^*(G)` by all parameters is finite dimensional.

Moreover, by [GreenKing]_, we are allowed to try and factorise the
simultanous lifts of the `d_{i,j}`, possibly after modification by
nilpotent elements. Let `D_i` be the smallest resulting non-constant
factor. And finally, the last parameter may be replaced by any other
parameter, since filter-regularity is automatic for the last parameter.

The motivation for all these refinements is to keep the parameter
degrees low. However, in the case of the dihedral group, we have `n=2`
and the Dickson class of `S_2` is of degree one, thus, is simple
enough.
 
If the Dickson classes for `G` have already been found, there is an
attribute ``Dickson``, which is a list of strings providing a
polynomial representation of the factors of Dickson classes. There are
two widely different ways to find the restricted Dickson elements, see
below. A list of parameters is returned by :meth:`~pGroupCohomology.cohomology.COHO.parameters`.

The dihedral group of order 8 has two non-conjugate maximal elementary
abelian subgroups of order 4. The restriction maps to these subgroups
can be obtained with ``H.restriction_maps()[2]`` and
``H.restriction_maps()[3]``. The restriction map to the greatest
central elementary abelian subgroup can be obtained with
``H.restriction_maps()[1]``.

We verify that the Dickson class of ``H`` really restricts to the
Dickson class of its two special subgroups of order 4, and that the
restriction to the greatest central elementary abelian subgroup is
zero. Note that the restricted Dickson classes of the subgroups are
not available if the optional parameter ``useElimination=True`` is
used (see below). In any case, the restricted Dickson classes of the
subgroups can be constructed using the method
:meth:`~pGroupCohomology.cohomology.COHO.dickson_in_subgroup`::

    sage: H.dickson_in_subgroup((4,2))
    sage: H.subgpDickson[4,2]
    [c_1_1: 1-Cocycle in H^*(SmallGroup(4,2); GF(2))]

During ``H.make(2)``, it was successfully tried to lift it to a
Dickson class in ``H``, using :meth:`~pGroupCohomology.cohomology.COHO.lift_dickson`.
The class is represented by a string::

    sage: H.Dickson
    ['b_1_0 + b_1_1']

There is an alternative way of lifting the Dickson classes. While
:meth:`~pGroupCohomology.cohomology.COHO.lift_dickson` works with
linear algebra, :meth:`~pGroupCohomology.cohomology.COHO.find_dickson`
uses elimination. The first method requires that the restriction maps
to the maximal elementary abelian subgroups are known out to the
degree in which the Dickson classes live. The degree grows
exponentially with the difference of the `p`-ranks of the group and its
center and can thus be very high, making this method
unfeasible. Hence, if the degree of the Dickson classes is greater
than 18, the elimination method is used by default. Using the optional
parameter ``useElimination=True``, the elimination method can be
chosen explicitly and can be explicitly avoided by
``useElimination=False``.

In the case of the dihedral group of order 8, both methods work, and
in fact they yield the same result (which is not necessarily the
case)::

    sage: H.Dickson = []
    sage: H.find_dickson()
    sage: H.Dickson
    ['b_1_0 + b_1_1']

We verify that the element ``H('b_1_1+b_1_0)`` restricts to zero in
the greatest central elementary abelian subgroup, and restricts to the
Dickson class in the two maximal elementary abelian subgroups::

    sage: print H.restriction_maps()[1][1](H('b_1_1+b_1_0'))
    1-Cocycle in H^*(SmallGroup(2,1); GF(2)),
    represented by
    [0]
    sage: H.subgpDickson[4,2][0] == H.restriction_maps()[2][1](H('b_1_1+b_1_0'))
    True
    sage: H.subgpDickson[4,2][0] == H.restriction_maps()[3][1](H('b_1_1+b_1_0'))
    True

Hence, in our example, we already know that if the cohomology
computation were complete then ``['c_2_2','b_1_0+b_1_1']`` would give
rise to a filter regular system of parameters. Filter regularity can
be checked as follows::

    sage: H.parameters()
    ['c_2_2', 'b_1_0 + b_1_1']
    sage: H.is_filter_regular(['c_2_2','b_1_0+b_1_1'], [2,1])
    True

Now, one considers the so-called (raw) filter-degree type. After
calling :meth:`~pGroupCohomology.cohomology.COHO.is_filter_regular`,
raw filter degree type and filter degree type are available by
::

    sage: H.rfdt
    [-1, -1, 1]
    sage: H.filter_degree_type()
    [-1, -2, -2]

The filter degree type together with the degrees of the filter regular 
system of parameters and the rank of the centre of `G` give rise to a 
degree bound of [Benson]_:

**Theorem**

  Let `G` be a finite group. Assume that `R_n` is an approximation of
  `H^\ast(G;\mathbb F_p)` out to degree `n` and assume that there is
  a filter regular HSOP for `R_n` in degrees at least two that gives
  rise to a filter regular HSOP of `H^\ast(G;\mathbb F_p)`. Let
  `f_0,...,f_r` be the filter degree type and let `d_1,...,d_r` be the
  degrees of the filter regular HSOP. If

    `n > \max(0, f_0+0,  f_1+1,...,f_{r-1}+r-1) - r + d_1+ ... +d_r` 

  then `R_n` is isomorphic to the cohomology ring.

Benson conjectures that the filter degree type for any modular
cohomology ring of a finite p-group is of the form
`-1,-2,...,-r,-r`. Note that the strict inequality in the theorem can
be improved to a non-strict inequality if the depth of the cohomology
ring of `G` is of rank at least two; in particular, by a result of
[Duflot]_, this is the case if the centre of a Sylow `p`-subgroup of
`G` is of rank at least two.

How do we apply that theorem? Apparently, the key problem is to find a
filter regular HSOP for the approximation `R_n` of the cohomology
ring. The point is that the Duflot regular generators together with
the Dickson classes form a filter regular HSOP for `H^\ast(G;\mathbb
F_p)`, and that we are able to compute Duflot regular generators and
Dickson classes in `R_n`.

1. Test whether the Duflot regular generators together with the Dickson
   classes form a filter regular HSOP for `R_n`. If it fails, then
   certainly `R_n` is not isomorphic to the cohomology ring. So, we
   need to compute an approximation in higher degrees.
2. If they *do* form a filter regular HSOP for `R_n` then we can compute
   the filter degree type and check whether the inequality in Benson's
   theorem is satisfied.
3. If it *is* satisfied, then `R_n` is isomorphic to the cohomology
   ring, and we can stop. Otherwise, we can not prove that `R_n` is
   isomorphic to the cohomology ring, and thus we must continue in
   higher degrees.

What does it mean, in our example? We have verified that we have 
a filter regular HSOP for the approximation in degree two. But looking
at the degrees of the parameters, Benson's criterion yields the degree 
bound 3. However, it turns out that there is no further generators or 
relations for ``H`` in degree 3. Hence, Benson's criterion is not 
perfect, as it sometimes applies only in higher degrees than necessary.

According to [GreenKing]_, it suffices to have one filter regular hsop
of the ring approximation that gives rise to a hsop of the cohomology
ring, and to prove the existence of a filter regular hsop of smaller
degrees by changing the coefficients to a finite extension field ---
then, one can use Benson's criterion using the degrees of the smaller
hsop (without explicitly constructing it). We elaborate this in the
next section.

Since the filter degree type is independent of the choice of a
filter-regular system of parameters, we can use the Dickson classes
for computing the filter-degree type. Then, we prove the existence of
a filter regular HSOP in smaller degrees. We do not need this HSOP
explicitly --- we only need the degrees (which comes out of the
existence proof) and the filter-degree type (which can be computed
using the Dickson classes).

Peter Symonds' Completeness Criterion
-------------------------------------

**Theorem** [Symonds]_

  Let `R_n` be the approximation of `H^\ast(G;\mathbb F_p)` out to
  degree `n`. Assume that there is a HSOP for `R_n` in degrees
  `d_1,...,d_r` that gives rise to a HSOP of `H^\ast(G;\mathbb
  F_p)`. Let `N` be the maximal degree of generators of `R_n` as a
  module over the parameters. If

    `n \ge \max(N, d_1+ ... +d_r - r +1)` 

  then `R_n` is isomorphic to the cohomology ring.

In order to compute the maximal degree `N` of module generators over
the parametes, it is needed to explicitly know the parameters. Hence,
in contrast to the Benson criterion, it seems impossible to improve
this criterion by an existence result for small parameters over an
extension field.

However, for verifying the Symonds criterion, it is not needed to
compute the filter degree type of the HSOP. Therefore, if our
construction methods happen to result in sufficiently small
parameters, we try to prove completeness based on the Symonds
criterion. Otherwise, we try to prove the existence of smaller
parameters over a finite extension field, as in the following section,
and apply a modification of the Benson criterion.


A step-by-step example
----------------------

The purpose of this section is to illustrate what happens behind the scenes
when invoking :meth:`~pGroupCohomology.cohomology.COHO.make`. Of course,
it is not needed to do all these steps manually when computing a cohomology
ring. 

... using a modification of the Benson criterion
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

In our first example, we use group number 32 of order 64.
::

    sage: H = CohomologyRing.user_db(64,32,websource=False)

We compute the ring structure out to degree 6. Since we did not use
the optional parameter ``useElimination=True``, some of the Dickson
classes of `G` will be found by the linear algebra method. For the
last one, this would be impossible by the linear algebra method, as
this Dickson class is of degree 7. Note that instead of repeating
``H.next()`` six times, one could also do ``H.make(6)``.
::

    sage: H.next()
    sage: H.next()
    sage: H.next()
    sage: H.next()
    sage: H.next()
    sage: H.next()
    sage: H.Dickson
    ['b_1_1^4 + b_2_1^2 + b_2_2^2 + b_2_3^2 + b_1_1*b_3_6',
     'b_2_2^2*b_1_1^2 + b_2_3^2*b_1_1^2 + b_1_1^3*b_3_6 + b_2_1*b_2_3^2 + b_2_2*b_2_3^2 + b_2_2*b_1_1*b_3_6',
     None]

Although the third Dickson class is in degree 7 and we know the ring
structure only out to degree 6, it is possible to express the third
Dickson class as a polynomial in the generators of our approximation
of the cohomology ring. This fact can not be seen with the linear
algebra method, but it can be seen using elimination, which is done in
:meth:`~pGroupCohomology.cohomology.COHO.find_dickson`.  Note that, by
default, it is tried to factorise the lift of the Dickson classes,
potentially after a modification by some nilpotent element (see
:meth:`~pGroupCohomology.cohomology.COHO.factorise`). By this little
trick, we actually obtain a parameter in degree one::

    sage: H.find_dickson()
    sage: H.Dickson
    ['b_1_1^4 + b_2_1^2 + b_2_2^2 + b_2_3^2 + b_1_1*b_3_6',
     'b_2_2^2*b_1_1^2 + b_2_3^2*b_1_1^2 + b_1_1^3*b_3_6 + b_2_2*b_2_3^2 + b_2_2*b_1_1*b_3_6 + b_3_5^2',
     'b_1_1']

Since the rank of the centre of the group is one, there is exactly one
Duflot regular generator.  We have already found it, namely in degree
4::

    sage: print H.5
    4-Cocycle in H^*(SmallGroup(64,32); GF(2)),
    represented by
    [0 0 0 0 0 0 0 0 0 0 0 1]
    rdeg = 1
    ydeg = 0
    
Now, as we have the Dickson classes and the Duflot regular generator,
does it make sense to try and apply Benson's criterion? It does not,
since there is one non-regular generator in degree 5::

    sage: H.gens()
    [1,
     b_2_1: 2-Cocycle in H^*(SmallGroup(64,32); GF(2)),
     b_2_2: 2-Cocycle in H^*(SmallGroup(64,32); GF(2)),
     b_2_3: 2-Cocycle in H^*(SmallGroup(64,32); GF(2)),
     b_4_10: 4-Cocycle in H^*(SmallGroup(64,32); GF(2)),
     c_4_11: 4-Cocycle in H^*(SmallGroup(64,32); GF(2)),
     a_1_0: 1-Cocycle in H^*(SmallGroup(64,32); GF(2)),
     b_1_1: 1-Cocycle in H^*(SmallGroup(64,32); GF(2)),
     b_3_4: 3-Cocycle in H^*(SmallGroup(64,32); GF(2)),
     b_3_5: 3-Cocycle in H^*(SmallGroup(64,32); GF(2)),
     b_3_6: 3-Cocycle in H^*(SmallGroup(64,32); GF(2)),
     b_5_17: 5-Cocycle in H^*(SmallGroup(64,32); GF(2))]

We thus expect that there will be a relation at least in degree
2*5=10. This heuristical lower bound for the necessary degree of
computations is available by a method::

    sage: H.expect_last_relation()
    10

So, we continue out to degree 10. Indeed, we find a relation in that
degree, but there are no further generators. Now it makes sense to
test whether the Duflot regular generator and the lifts of the Dickson
classes form a filter regular homogeneous system of parameters::

    sage: H.next()
    sage: H.next()
    sage: H.next()
    sage: H.next()
    sage: H.lastRel
    10
    sage: H.is_filter_regular(['c_4_11']+H.Dickson, [4,4,6,1])
    True

The Filter Degree Type is in accordance with Benson's Strong
Regularity Conjecture::
    
    sage: H.filter_degree_type()
    [-1, -2, -3, -4, -4]

The original form of Benson's criterion would only apply in a very high degree:

- The `p`-rank of the group is 4. Benson originally suggested to construct
  parameters by lifting Dickson elements in the maximal elementary abelian
  subgroups (*not* restricted to a complement of the greatest central
  elementary abelian subgroup). These unrestricted Dickson classes live
  in degrees 8, 12, 14 and 15, or powers thereof.
- The centre is of rank less than two, so, we can not be sure that the
  depth of the cohomology ring is at least two.
- By consequence, Benson's criterion applies, in the best case, in
  degree `7+11+13+14+1=46`.
 
However, by our improved choice of filter-regular parameters, we obtain
the degree bound `3+3+5+1+1=13`. But we can still do better.

**Theorem** [GreenKing]_

    Let `p_1,...,p_r` be a filter-regular HSOP for the approximation
    `R_n` of `H^\ast(G;\mathbb F_p)` that gives rise to a filter
    regular HSOP for the cohomology ring. Assume that for some `m\le r`,
    the quotient of `R_n` modulo `p_1,...,p_m` and modulo all monomials
    of some degree `d` is finite.

    Then, there is a filter-regular HSOP for the degree `n`
    approximation of `H^\ast(G;K)` for some finite field extension
    `K` of `\mathbb F_p`, obtained from `p_1,...,p_m` and `r-m`
    elements of degree `d`.

    Moreover, the degrees of this HSOP of `H^\ast(G;K)` can be used
    as in Benson's criterion to prove completeness of the ring
    approximation of `H^\ast(G;\mathbb F_p)`.

This result is used in
:meth:`~pGroupCohomology.cohomology.COHO.potential_degree_bound`::

    sage: H.potential_degree_bound(['c_4_11']+H.Dickson, [4,4,6,1])
    7

That means: If it is true that ``['c_4_11']+H.Dickson`` is a
filter-regular HSOP of ``H`` of an appropriate filter degree type and
if we happen to know the ring structure is known at least out to
degree 7, then the computation is complete. Since we already verified
filter regularity and since we know the ring structure not only to
degree 7 but to degree 10, we are done.

We remark that we only need to know the mere *existence* of a
filter-regular HSOP in low degrees *over a finite field extension*.
It is not needed, and in general impossible, to construct the small
parameters without extending the field. However, here it is possible,
and we do so, just for illustration. 

The attribute ``WhatFRS`` tells us that we can replace the last three 
parameters (which are of degree 4, 6, and 7) by elements in degree 2::

    sage: H.WhatFRS
    (3, 2)

Since usually we would only work with a single filter-regular HSOP and
since the test for filter-regularity is expensive, we cache the
result. So, before testing another HSOP, we have to clear the
cache. Then we see whether the three generators in degree two can be
used as parameters::

    sage: H.delprop('lastTested')
    sage: H.is_filter_regular(['c_4_11','b_2_1','b_2_2','b_2_3'], [4,2,2,2])
    False

This is not the case, so, we try some other elements::

    sage: H.delprop('lastTested')
    sage: H.is_filter_regular(['c_4_11', 'b_2_1+b_2_2', 'b_2_1+b_2_3', 'b_2_2+b_2_3'], [4,2,2,2])
    False
    sage: H.delprop('lastTested')
    sage: H.is_filter_regular(['c_4_11', 'b_2_1+b_2_2', 'b_1_1^2+b_2_3', 'b_2_2+b_2_3'], [4,2,2,2])
    True

So, indeed, it is possible (but not needed) to explicitly provide
three filter regular parameters in degree two, in addition to the
Duflot regular generator.

... using the Symonds criterion
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

**Theorem** [Symonds]_

    Assume that the cohomology ring is known out to degree `n`,
    and assume that the ring approximation has a hsop `P` in 
    degrees `d_1,...,d_r` that gives rise to a hsop of the 
    cohomology ring. If `n > \sum_{i=1}^r (d_i-1)` and if the 
    ring approximation is generated in degree at most `n` as 
    a module over the algebra generated by `P`, then the ring
    approximation is complete.

Note that it is not assumed that `P` is filter regular. This gives
more freedom. The strategy in :meth:`~pGroupCohomology.cohomology.COHO.parameters`
is to start with a filter regular hsop, which can be constructed
using Duflot generators and Dickson elements as explained above,
see :meth:`~pGroupCohomology.cohomology.COHO.filter_regular_parameters`.
Then, it is tested by exhaustive search whether any of these
filter regular parameters can be replaced by a parameter of smaller 
degrees. This will, in general, result in a hsop that is not 
filter regular, but has smaller degrees.

In our example, we indeed obtain parameters in smaller degrees::

    sage: H.filter_regular_parameters()
    ['c_4_11', 'b_1_1^4 + b_2_1^2 + b_2_2^2 + b_2_3^2 + b_1_1*b_3_6', 'b_2_2^2*b_1_1^2 + b_2_3^2*b_1_1^2 + b_1_1^3*b_3_6 + b_2_2*b_2_3^2 + b_2_2*b_1_1*b_3_6 + b_3_5^2', 'b_1_1']
    sage: H.parameters()
    ['c_4_11', 'b_2_3+b_2_2+b_2_1', 'b_2_2+b_2_1', 'b_1_1']

In this particular case, the small hsop is filter regular, but
we can not rely on it. Anyway, we have `10>4+2+2+1-4`, and it
turns out that the Symonds test applies, since in fact the
ring approximation is generated in degree at most 5 over the
parameters::

    sage: H.SymondsTest(H.parameters(),[4,2,2,1])
    True
    sage: H._max_module_deg
    5

There are cases in which the test for filter regularity can be
quite expensive. On the other hand, it can be very expensive to
compute a minimal resolution in needlessly high degree. Hence,
*if* Symonds' criterion has a chance to apply then we try this
first. The modified Benson test is only used if it would yield 
a better degree bound, due to the existence of smaller parameters
over an extension field. Here, the Symonds test is used. This
can be seen in the protocol (after resetting an internal parameter)::

    sage: H.suffDeg=-1
    sage: H.option('prot')
    sage: H.test_for_completion()
    Trying Symonds' criterion
    Complete Groebner basis of H^*(SmallGroup(64,32); GF(2)) is already known
    Successful application of the Symonds criterion
    True
    sage: H.option('noprot')


3. Background for the Cohomology Computation of Non Prime Power Groups
======================================================================

In this section, we explain the computation of the modular cohomology
of a finite group that is not of prime power order. We do it by means
of an example that we compute step by step. Of course, invoking
:meth:`~pGroupCohomology.modular_cohomology.MODCOHO.make` will do all
these computations behind the scenes.

Our main example is the mod-2 cohomology of the Symmetric Group on
8 elements. We use the option ``useFactorization=False``, which
means that it is not attempted to decompose the parameters into 
factors. The reason is that this factorisation (depending on the
state of the randomness generator) may take a long time.
::

    sage: G = gap('SymmetricGroup(8)')
    sage: tmp = tmp_filename()
    sage: CohomologyRing.set_user_db(tmp)
    sage: HS8 = CohomologyRing(G, prime=2, GroupName='SymmetricGroup(8)', useFactorization=False)


The Stable Element Method
-------------------------

For the theoretical background of the Stable Element Method, we refer to [AdemMilgram]_.
It is known that the mod-`p` cohomology ring of a finite group `G` is
a sub-ring of the mod-`p` cohomology of any subgroup `P` whose index in `G`
is co-prime to `p`. This sub-ring is formed by those cocycles of `P`
that are stable under conjugation by elements of `G`. That means, a
cocycle `c\in H^\ast(P; \mathbb F_p)` is stable if and only if for all
`g\in G` holds `r_{P, P\cap P^g}(c) = r_{P^g, P\cap P^g}(\phi_g(c))`,
where `\phi_g` denotes the homomorphism induced by conjugation with `g^{-1}`,
and `r_{P, P\cap P^g}, r_{P^g, P\cap P^g}` are restriction maps.

It suffices to test stability for representatives of all double cosets of `P`
in `G`. While this is, of course, much better than testing all elements of `G`,
the number of double cosets can still be rather large. It is thus often
better to compute the cohomology of `G` in two steps. By default, we proceed as
follows.

 - Let `S` be a Sylow `p`-subgroup of `G`. Compute its mod-`p` cohomology,
   ``HS``.
 - Let `N = N_G(Z(S))`. Compute its mod-`p` cohomology ``HN`` as a sub-ring
   of ``HS``.
 - If `N\not=G`: Compute the mod-`p` cohomology of `G` as a sub-ring of ``HN``.
    
In our example, the subgroups look as follows::

    sage: HS8.subgroup_cohomology() # HS8 is a sub-ring of HS8.subgroup_cohomology()
    H^*(SmallGroup(384,5602); GF(2))
    sage: HS8.subgroup_cohomology().subgroup_cohomology()
    H^*(Syl2(S8); GF(2))
    sage: HS8.subgroup_cohomology().subgroup_cohomology() == HS8.sylow_cohomology()
    True

The stability conditions are expressed by pairs of induced
homomorphisms starting at ``HS8.subgroup_cohomology()``::

    sage: HS8._PtoPcapCPdirect
    [Induced homomorphism of degree 0 from H^*(SmallGroup(384,5602); GF(2)) to H^*(D8xV4; GF(2)), Induced homomorphism of degree 0 from H^*(SmallGroup(384,5602); GF(2)) to H^*(SmallGroup(32,27); GF(2)), Induced homomorphism of degree 0 from H^*(SmallGroup(384,5602); GF(2)) to H^*(D8; GF(2))]
    sage: HS8._PtoPcapCPtwist
    [Induced homomorphism of degree 0 from H^*(SmallGroup(384,5602); GF(2)) to H^*(D8xV4; GF(2)), Induced homomorphism of degree 0 from H^*(SmallGroup(384,5602); GF(2)) to H^*(SmallGroup(32,27); GF(2)), Induced homomorphism of degree 0 from H^*(SmallGroup(384,5602); GF(2)) to H^*(D8; GF(2))]

The elements of ``HS8`` are given by those elements of ``HS8.subgroup_cohomology()``
for which the images coincide for each pair. These so-called *stable elements*
can be computed in a specific degree by
:meth:`~pGroupCohomology.modular_cohomology.MODCOHO.stable_space`,
which also returns a list of standard monomials for ``HS8.subgroup_cohomology()``
and a sublist of those monomials that occur as leading monomials in
stable cocycles::

    sage: HS8.stable_space(2)
    ([Gen2: 2-Cocycle in H^*(SymmetricGroup(8); GF(2)), Gen1: 2-Cocycle in H^*(SymmetricGroup(8); GF(2))], ['b_1_1^2', 'b_1_0*b_1_1', 'b_1_0^2', 'b_2_4', 'b_2_3'], ['b_1_1^2', 'b_1_0*b_1_1'])


Construction of filter regular parameters
-----------------------------------------

In the case of prime power groups, we form a filter regular hsop out
of a Duflot regular sequence (choosing the parameters such that a
Duflot regular sequence is formed by generators), together with
simultaneous lifts of Dickson elements in the maximal elementary
abelian subgroups in a complement of the greatest central elementary
abelian subgroup. For general finite groups, we need to use slightly
different methods.

Duflot regular sequence
^^^^^^^^^^^^^^^^^^^^^^^

In our main example ``HS8``, the rank of the centre of a Sylow
subgroup is one. Hence, any Duflot element yields a maximal Duflot
regular sequence. This is found in degree four. Note that the second
index of the variable name is somehow random.
::

    sage: HS8.make(4)
    sage: HS8.duflot_regular_sequence()
    ['c_4_...']

There are some difficulties for non prime power groups, though. For
starters, the number of generators with non-trivial restriction to the
greatest central elementary abelian subgroup of a Sylow `p`-subgroup
exceeds the rank of the greatest central `p`-elementary abelian
subgroup. In the following example, all 5 generators are Duflot
elements, but only three of them form a regular sequence::

    sage: H96 = CohomologyRing(96,3,prime=2)
    sage: H96.make()
    sage: [t for t in H96.gens()[1:] if t.rdeg()]
    [c_2_0: 2-Cocycle in H^*(SmallGroup(96,3); GF(2)), c_4_3: 4-Cocycle in H^*(SmallGroup(96,3); GF(2)), c_6_4: 6-Cocycle in H^*(SmallGroup(96,3); GF(2)), c_6_6: 6-Cocycle in H^*(SmallGroup(96,3); GF(2)), c_6_7: 6-Cocycle in H^*(SmallGroup(96,3); GF(2))]
    sage: H96.duflot_regular_sequence()
    ['c_2_0', 'c_4_3', 'c_6_7']

Note that sometimes the program will only find to a Duflot regular
sequence that is not entirely formed by generators, as in the
following example::

    sage: H48 = CohomologyRing(48,50, prime=2)
    sage: H48.make()
    sage: H48.duflot_regular_sequence()
    ['c_2_3', 'c_2_2', 'c_3_1', 'c_3_7+c_3_0']


Simultaneous lifts of Dickson elements
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

By [Benson]_, it is always possible to find elements in the cohomology
of `G` that restrict to (powers of) Dickson elements in all maximal
elementary abelian subgroups, and these elements form a filter regular
hsop.

The degrees of these elements grow exponentially in the `p`-rank of
`G`. This is not good, since all known completeness criteria depend on
the degrees of the parameters. However, Benson's construction
certainly is a reasonable starting point, that we refine by various
methods.

Dickson elements in a complement of the centre of a Sylow subgroup
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Above, we stated that a filter regular hsop results if one extends a
maximal Duflot regular sequence by simultaneous lifts of Dickson
elements of complements of the greatest central elementary abelian
subgroup of a Sylow subgroup in maximal elementary abelian
subgroups. This holds in general.

However, the *existence* of the simultaneous lifts is only granted
for prime power groups [GreenKing]_. For groups that are not of
prime power order, they do not exist, in general.

But often enough, the simultaneous lifts *do* exist. Therefore, we
proceed as follows.

    * Since we use the stable element method, the cohomology ring
      of a Sylow subgroup `S` is already known when we start the 
      cohomology computation of a finite group `G`.
    * Simultaneous lifts of Dickson elements in maximal elementary 
      abelian subgroups in a complement of the greatest central 
      elementary abelian subgroup of `S` exist in the cohomology
      of `S`.
    * If these simultaneous lifts happen to be stable, then they
      yield elements of the cohomology of `G`. Together with a
      maximal Duflot regular sequence of the cohomology of `G`,
      they yield a filter regular hsop.

In the following example, the `p`-rank of the group is four, hence,
the Dickson elements are in degrees 8, 12, 14 and 15. But the
centre of a Sylow subgroup is of rank two, and therefore, the
Dickson elements in a complement of the centre of a Sylow subgroup
are of degree 2 and 3. In fact, the degree 3 element can even
be replaced by a parameter in degree 1, for the Sylow subgroup::

    sage: H576 = CohomologyRing(576,5000, prime=2)
    sage: PS = H576.sylow_cohomology().filter_regular_parameters(); PS
    ['c_4_50', 'c_1_4', 'b_1_0^2 + b_1_0*b_1_1 + b_1_1^2 + b_1_0*b_1_2 + b_1_2^2 + b_1_1*b_1_3 + b_1_2*b_1_3 + b_1_3^2', 'b_1_1 + b_1_2']

Since some algorithms of Gap used in the background are randomized, it
may or may not be that all chosen possibly factorised restricted
Dickson elements are stable::

    sage: H576.make()
    sage: H576.parameters_from_sylow_subgroup() #random
    ['c_1_2', 'c_4_18', 'b_1_1^2 + b_1_1*b_1_0 + b_1_0^2', 'b_1_1']

If the last parameter for the cohomology of the Sylow subgroup were not
stable, we could use :meth:`~pGroupCohomology.cohomology.COHO.find_small_last_parameter`
to find another parameter, which is done internally in
:meth:`~pGroupCohomology.cohomology.COHO.filter_regular_parameters`.

For ``H96`` (an example that we started above), the rank of a maximal
2-elementary abelian subgroup is three. So, the Dickson elements are
of degree 4, 6 and 7. This would only give rise to a degree bound of
3+5+6=14. Nevertheless, the Duflot regular sequence provides a better
filter-regular HSOP, and indeed the program automatically replaces the
Dickson elements by the Duflot generators::

    sage: H96.filter_regular_parameters()
    ['c_2_0', 'c_4_3', 'c_6_7']

Just for demonstration, we reconstruct the original Dickson elements,
without any simplification::

    sage: H96.setprop('useFactorization',False)
    sage: H96.find_dickson()
    sage: H96.Dickson
    ['c_4_3^2+c_2_0*c_6_7+c_2_0^2*c_4_3+c_2_0^4', 'c_6_4*c_6_7+c_6_4^2+c_4_3^3+c_2_0^2*c_4_3^2', 'c_2_0*c_6_4*c_6_7+c_2_0*c_6_4^2+c_2_0*c_4_3^3+c_2_0^4*c_6_7+c_2_0^5*c_4_3']


Finding smaller parameters
^^^^^^^^^^^^^^^^^^^^^^^^^^

In [GreenKing]_, we show that the last element of a filter regular
hsop can always be replaced by any other parameter --- the result will
still be a filter regular hsop. The method :meth:`~pGroupCohomology.cohomology.COHO.find_small_last_parameter`
finds a parameter of minimal degree, by exhaustive search.

However, if this replacement is not done in the last position, the
resulting hsop is not necessarily filter regular. But both for the
Symonds and the Hilbert-Poincar\'e criterion explained below, a
general hsop suffices. In the following example, a hsop can be
constructed in much smaller degrees than a filter regular hsop.
Since some randomised algorithms of Gap are used in the background,
we only give the degrees of the parameters::

    sage: H = CohomologyRing(384,18136,prime=2)
    sage: H.make()
    sage: H.set_ring()
    sage: [singular('deg(%s)'%p) for p in H.filter_regular_parameters()]
    [8, 12, 14, 3]
    sage: [singular('deg(%s)'%p) for p in H.parameters()]
    [4, 4, 3, 2]

These parameters are not filter regular::

    sage: H.is_filter_regular(H.parameters(),[4,4,3,2])
    False

We return to our main example ``HS8``. In this example, the parameters
for the cohomology of the Sylow subgroup are not stable. The last Dickson
element for ``HS8`` is of degree 15. But we don't need to compute that far,
since the Dickson elements can be expressed by generators in smaller
degrees. Again, the variable names are somehow random.
::

    sage: HS8.filter_regular_parameters()
    sage: HS8.make(7)
    sage: HS8.parameters_from_sylow_subgroup()
    ['c_4_...', None, None, None]
    sage: HS8.option('prot')
    sage: HS8.generator_degbound()
    New generators for H^*(SymmetricGroup(8); GF(2)) will be found at most in degree 8
    sage: HS8.option('noprot')
    sage: HS8.find_dickson()
    sage: P = HS8.Dickson
    sage: HS8.set_ring()
    sage: [singular('deg(%s)'%p) for p in P]
    [8, 12, 14, 15]

As the ring approximation is carried to higher degrees, we can eventually
replace the last parameter by an element of degree 6::

    sage: HS8.make(14)
    sage: HS8.find_small_last_parameter(P)
    True
    sage: [singular('deg(%s)'%p) for p in P]
    [8, 12, 14, 6]

The above construction is used in :meth:`~pGroupCohomology.cohomology.COHO.filter_regular_parameters`,
and so we obtain
::

    sage: P == HS8.filter_regular_parameters()
    True

The modified Benson criterion will only apply in rather high degree::

    sage: HS8.potential_degree_bound(P,[8,12,14,6])
    36

By exhaustive search, we find that we can replace the first parameter
by a parameter in degree 4. In fact, it is guaranteed that no element
of the following hsop can be replaced by an element of smaller
degree. However, there is no guarantee that this hsop is filter
regular, so that it is of no use for the Benson criterion.
::

    sage: [singular('deg(%s)'%p) for p in HS8.parameters()]
    [4, 12, 14, 6]

The Symonds criterion does not rely on filter regularity, and using
the above parameters it can apply in degree `4+12+14+6-3=33`.  However,
another criterion works still better, as we explain in the following
section.


A completeness criterion for non prime power groups
---------------------------------------------------

Above, we explained how Benson's and Symonds' completeness criteria
work. For non prime power groups, a criterion is available that
combines the advantages of both. It is formed by two separate
completeness criteria for generators and relations. See [King]_
for more details.

Bounding the generator degree
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

It is well known that `H^\ast(P; \mathbb F_p)` is a finitely generated
`H^\ast(G; \mathbb F_p)`-module via the restriction from `G` to `P`.
Using the fact that transfer is a module homomorphism, one can easily
show that the maximal degree of a generator of `H^\ast(G; \mathbb F_p)`
is bounded by the generator degree of `H^\ast(P; \mathbb F_p)` as a
`H^\ast(G; \mathbb F_p)`-module. This can also be computed in an
approximation of `H^\ast(G; \mathbb F_p)`, and is implemented in
:meth:`~pGroupCohomology.modular_cohomology.MODCOHO.generator_degbound`.

This method was automatically invoked when we computed the ring structure
out to degree 14. Two attributes provide the result::

    sage: HS8.degbound_for_gens
    8
    sage: HS8.all_generators_found
    True

Hence, the program knows that all stable elements above degree 8 can be 
expressed as polynomials of the generators of the ring approximation. It 
is then not needed to compute the stable subspace by means of 
:meth:`~pGroupCohomology.modular_cohomology.MODCOHO.stable_space`, which 
saves a considerable amount of resources.

The Hilbert-Poincare criterion
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

There remains to determine whether all relations have been found.
The following result arose in a discussion of Simon King with
Peter Symonds. Recall that we compute `H^\ast(G; \mathbb F_p)`
as a subring of `H^\ast(P; \mathbb F_p)` for a subgroup `P<G`.

**Theorem**

    Assume that there is a finite field extension `K` over `\mathbb F_p`
    so that `H^\ast(G; K)` has a hsop `P` of degrees
    `d_1,...,d_r`. Let `N = \sum_{i=1}^r d_i - \mathrm{depth}H^\ast(P; \mathbb F_p)`.
    Let `R` be the degree-`n` approximation of `H^\ast(G; \mathbb F_p)`,
    with `n\ge N`, and assume that it contains all generators of
    `H^\ast(G; \mathbb F_p)`. 

    Then, `R` is isomorphic to `H^\ast(G; \mathbb F_p)`
    if and only if the Poincar\'e series of `R` multiplied with
    `\prod_{i=1}^r(1-t^{d_i})` is a polynomial of degree `\le N`.

We explained above how parameters for `H^\ast(G; \mathbb F_p)` can be
constructed in a ring approximation. We find parameterst in degrees 4,
12, 14 and 6. Let us test whether we can show the existence of smaller
parameters over a finite extension field::

    sage: HS8.option('prot')
    sage: HS8.HilbertPoincareTest()
    We expect that the Hilbert-Poincare criterion will not apply before degree 26
    sage: HS8.WhatHSOP
    ['4', '12', '6', '7']
    sage: HS8.option('noprot')
    sage: HS8.subgroup_cohomology().depth()
    3

The attribute ``WhatHSOP`` tells that it is possible to replace the
degree-14 parameter by a parameter in degree 7, after extending the
field. Since the depth of the cohomology of the subgroup is three,
there is a chance for the criterion to work in degree `26 = 4+12+6+7-3`.
We now complete the computation and verify the criterion manually::

    sage: HS8.make(26)
    sage: p = HS8.poincare_series()
    sage: t = p.parent().gen(0)
    sage: p*(1-t^4)*(1-t^12)*(1-t^6)*(1-t^7)
    t^25 + t^24 + 2*t^23 + 4*t^22 + 5*t^21 + 8*t^20 + 10*t^19 + 12*t^18 + 16*t^17 + 18*t^16 + 19*t^15 + 22*t^14 + 21*t^13 + 21*t^12 + 22*t^11 + 19*t^10 + 18*t^9 + 16*t^8 + 13*t^7 + 11*t^6 + 8*t^5 + 5*t^4 + 4*t^3 + 2*t^2 + t + 1

Indeed, the result is a polynomial of degree at most 26. Therefore, the
computation is complete.

Comparison of the criteria
^^^^^^^^^^^^^^^^^^^^^^^^^^

For applying the Symonds criterion, it is essential to have parameters
explicitly, so that one can compute the maximal degree of module generators.
Here, we explicitly find parameters in degrees 4, 12, 14 and 6. This means
that the Symonds criterion has only a chance to apply in degree 
`4+12+14+6-3=33`. Note that it is impossible to replace the degree-14
parameter by a parameter in degree 7 *without* extending the field, for
otherwise it would have been found in the exhaustive search that was
carried out in :meth:`~pGroupCohomology.cohomology.COHO.parameters`.

The best filter regular hsop that our automatic methods can provide
has degrees 8, 12, 14 and 6::

    sage: [singular('deg(%s)'%p) for p in HS8.filter_regular_parameters()]
    [8, 12, 14, 6]

It turns out that even in an extension field it is not possible to
replace the last few parameters in smaller degrees. Hence, the
modified Benson test only applies in degree `8+12+14+6-4=36`.

The Hilbert-Poincar\'e criterion combines the advantage of the
modified Benson criterion (the existence result for small parameters
over an extension field) with the advantage of the Symonds criterion
(the conditions in the criterion are easy to verify). Hence, we use it
as the default criterion for non prime power groups.

References
==========

.. [AdemMilgram] Alejandro Adem, R. James Milgram: *Cohomology of finite groups.* Second edition. Grundlehren der Mathematischen Wissenschaften [Fundamental Principles of Mathematical Sciences], **309.** Springer-Verlag, Berlin, 2004.
.. [Benson] Dave Benson: Dickson invariants, regularity and computation in group cohomology. *Illinois J. Math.* **48** (2004), 171--197.
.. [BensonCarlson] Dave Benson, Jon F. Carlson: Projective resolutions and Poincar\'e duality complexes. *Trans. Amer. Math. Soc.* **342** (1994), 447--488.
.. [CartanEilenberg] Henri Cartan and Samuel Eilenberg: *Homological algebra*. Princeton University Press, Princeton, N. J., 1956.
.. [Duflot] Jeanne Duflot: Depth and equivariant cohomology, *Comment. Math. Helv.* **56** (1981), 627--637.
.. [Green] David Green: Gr\"obner bases and the computation of group cohomology.  *Lecture Notes in Mathematics,* **1828**. Springer-Verlag, Berlin, 2003.
.. [GreenKing] David Green, Simon King: `The computation of the cohomology rings of all groups of order 128. <http://arxiv.org/abs/1001.2577>`_, *Preprint,* 2010.
.. [King] Simon King: `Completeness criteria for modular cohomology rings of non prime power groups. <http://arxiv.org/abs/1004.0736>`_, *Preprint,* 2010.
.. [Kraines] David Kraines: Massey Higher Products. *Trans. AMS* **124** (1966), 431--449.
.. [Kuhn] Nicholas J. Kuhn: Primitives and central detection numbers in group cohomology, *Adv. Math.* **216** (2007), 387--442.
.. [Pham] Pham Anh Minh: Modular invariant theory and cohomology algebras of extra-special `p`-groups. *Pacific J. Math.* **124** (1986), 345--363.
.. [Ravenel] Douglas Ravenel, *Complex cobordism and stable homotopy groups of spheres,* Pure and Applied Mathematics, **121**. Academic Press, Inc., Orlando, FL, (1986).
.. [Symonds] Peter Symonds: `On the Castelnuovo-Mumford Regularity of the Cohomology Ring of a Group <http://www.maths.manchester.ac.uk/~pas/preprints/>`_.  Preprint, June 2009. 
.. [Wilkerson] Clarence Wilkerson: *A primer on the Dickson invariants,* Proceedings of the Northwestern Homotopy Theory Conference (Evanston, Ill., 1982), *Contemp. Math.,* **19** (1983), 421--434.


"""

from sage.all import SAGE_ROOT, DOT_SAGE, load
from sage.rings.integer import Integer
from sage.interfaces.singular import singular
from pGroupCohomology.cohomology import COHO
from pGroupCohomology.resolution import OPTION
from pGroupCohomology import barcode
import re

###########
## Cohomology rings will be unique parent structures
from weakref import WeakValueDictionary
_cache = WeakValueDictionary({})


##########
## A little regular expression that transforms any string into a valid GStem

_GStemMaker = re.compile(r'[^0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ]')

##########
## Transformation into latex
_index_match = re.compile('(_\d+)+')
_exp_match = re.compile('\^\d')
_name2latex = lambda t: _exp_match.sub(lambda m: '^{'+m.group()[1:]+'}', _index_match.sub(lambda m:'_{%s}'%','.join(m.group().split('_')[1:]),t).replace('**','^')).replace('*',' ')

class CohomologyRingFactory:
    r"""
    A factory for creating modular cohomology rings of finite p-groups as unique parent structures

    Please use :func:`~pGroupCohomology.CohomologyRing`, which is an
    instance of this class, and is provided with a documentation of
    the available options.
    
    TESTS::
    
        sage: from pGroupCohomology import CohomologyRing
        sage: H0 = CohomologyRing(8,3)
        sage: print H0
        Cohomology ring of Dihedral group of order 8 with coefficients in GF(2)
        <BLANKLINE>
        Computation complete
        Minimal list of generators:
        [c_2_2: 2-Cocycle in H^*(D8; GF(2)),
         b_1_0: 1-Cocycle in H^*(D8; GF(2)),
         b_1_1: 1-Cocycle in H^*(D8; GF(2))]
        Minimal list of algebraic relations:
        [b_1_0*b_1_1]
    
    """
    def __call__ (self, *args, **kwds):
        """
        Create the mod-p cohomology ring of a finite groups

        Of course, isomorphic p-Groups have isomorphic cohomology
        rings.  However, the presentation of the cohomology rings as
        obtained by our package depends on the choice of a minimal
        generating set of the p-group.
        
        If a `p`-group `G` is given by its position in the SmallGroups
        library, then this position, perhaps together with a custom
        name provided by the user, forms a unique key for the
        cohomology ring.
        
        If `G` is given as a group in the Gap interface, then it is
        required that the first items on the list of generators of `G`
        forms a minimal generating set. If this is not the case, an
        error is raised. We transform `G` into a permutation group
        whose generators correspond to a minimal generating set of
        `G`. The description of that permutation group, perhaps
        together with a custom name, forms a unique key for the
        cohomology ring.
        
        The unique key also depends on the chosen folders containing
        data of the ring.

        TESTS::

            sage: from pGroupCohomology import CohomologyRing
            sage: H0 = CohomologyRing(8,3) # indirect doc test
            sage: print H0
            Cohomology ring of Dihedral group of order 8 with coefficients in GF(2)
            <BLANKLINE>
            Computation complete
            Minimal list of generators:
            [c_2_2: 2-Cocycle in H^*(D8; GF(2)),
             b_1_0: 1-Cocycle in H^*(D8; GF(2)),
             b_1_1: 1-Cocycle in H^*(D8; GF(2))]
            Minimal list of algebraic relations:
            [b_1_0*b_1_1]
            sage: tmp_root = tmp_filename()
            sage: H1 = CohomologyRing(8,3,root=tmp_root)
            
        """
        from pGroupCohomology.cohomology import COHO
        from pGroupCohomology.modular_cohomology import MODCOHO
        from pGroupCohomology.resolution import OPTION
        # Basic idea: 
        # The key shall both be a unique pointer to the data in the file 
        # system and a descriptor of the group-with-minimal-generators.
        # Hence, it is the root directory plus the stem name plus [either
        # the position in the SmallGroups library or a permutation group
        # presentation].
        # The GroupName and other properties are extra arguments

        if len(args)>2 or len(args)==0:
            raise ValueError, "The p-Group must be described by one or two parameters"

        # If cohomology options are required, they are provided now.
        # Note that these are valid for any subsequent computations with
        # any cohomology ring: The options are not associated with the
        # ring that we are returning below.
        if kwds.has_key('options'):
            from pGroupCohomology.cohomology import COHO_option
            if isinstance(kwds['options'], basestring):
                COHO_option(kwds['options'])
            else:
                COHO_option(*kwds['options'])

        pr = None
        if len(args) == 2:
            from pGroupCohomology.resolution import OPTION, gap
            gap.eval('RestoreStateRandom(0)')
            q = Integer(args[0])
            if not q.is_prime_power():
                pr = kwds.get('prime')
                if pr is None:
                    raise ValueError, "The parameter `prime` must be provided"
                try:
                    if not pr.is_prime():
                        raise ValueError
                except:
                    raise ValueError, "The parameter `prime` must provide a prime number"                    
            n = Integer(args[1])
            max_n = Integer(gap('NumberSmallGroups(%d)'%q))
            if not 0<n<=max_n:
                raise ValueError, "Second parameter must be between 1 and %d"%max_n
            KEY = (q,n)
            if pr is not None: # non prime power case
                Hfinal = gap('SmallGroup(%d,%d)'%(q,n))
            GStem = '%dgp%d'%(q,n)      # this should be chosen in an automated way
            GroupName = kwds.get('GroupName', None)  # this may be provided by the user.
                                                    # If the name is None, then it will be fixed
                                                    # in COHO.__init__
            # Setting the random seed is important, since finding
            # generating sets with gap is (to some extent) randomised.
            # But we need a computationally unique result.
        else:
            H0 = args[0]
            if not hasattr(H0,'parent'):
                raise TypeError, "We expected a group defined in Gap"
            gap = GAP = H0.parent()
            gap.eval('RestoreStateRandom(0)')
            # Setting the random seed is important, since finding
            # generating sets with gap is (to some extent) randomised.
            # But we need a computationally unique result.
            ####
            ## make up the (stem) name
            if GAP.eval('HasName(%s)'%(H0.name())) == 'true':
                GroupName = GAP.eval('Name(%s)'%(H0.name()))[1:-1]
            else:
                GroupName = None
            GroupName = kwds.get('GroupName', GroupName)
            GStem = kwds.get('GStem', GroupName)    # here, a unique short descriptor should be given -- up to the user,
                                                    # since we don't see a clear automated procedure of choice.
            if (GStem is None):
                if not kwds.get('GroupId'):
                    raise ValueError, "'GroupName' must be provided"
                q_,n_ = kwds['GroupId']
                GStem = '%dgp%d'%(q_,n_)
            if not isinstance(GStem, basestring):
                raise TypeError, "'GroupName' must be a string"

            ## Now the name is fixed (and can be any string).
            ## The GStem must be usable as a valid file name:
            GStem = _GStemMaker.sub('_',GStem)            

            ####
            ## process the gap input:
            #   1. good group order?
            #   2. minimal generators given?
            #      if not, find them!
            #   3. transform into permutation group
            if OPTION.opts['prot']:
                print "Computing group order"
            q = Integer(GAP.eval('Order(%s)'%(H0.name())))
            if OPTION.opts['prot']:
                print "The group is of order",q
            if not q.is_prime_power():
                pr = kwds.get('prime')
                if pr is None:
                    raise ValueError, "The parameter `prime` must be provided"
                try:
                    if not pr.is_prime():
                        raise ValueError
                except:
                    raise ValueError, "The parameter `prime` must provide a prime number"

            ## Try to make up minimal generators
            if kwds.get('minimal_generators'): 
                # We already have minimal generators (or the user thinks so...), and then we are supposed to use exactly *these* generators!
                Hfinal = H0
            else:
                if q.is_prime_power(): # otherwise, the minimal generating set thing is not so relevant
                    if OPTION.opts['prot']:
                        print "\n"+"Trying to find minimal generators for", GroupName or GStem
                    PhiP = H0.admissibleGroup()
                    if repr(PhiP)!='fail':
                        Hfinal = PhiP.Range()
                    else:
                        raise ValueError, "The first generators of %s must form a minimal generating set"%(H0)
                else:
                    Hfinal = H0
                    
            ####
            ## Finally, make up a description of the group which preserves the generators
            ## -- unless the user wants a different key (which may fail badly, but this is not our problem... :)
            # It may happen that the group-with-generators is equivalent to a SmallGroup:
            try:
                GId = kwds.get('GroupId') or Hfinal.IdGroup().sage()
                if GAP.eval('canonicalIsomorphism(%s,SmallGroup(%d,%d))'%(Hfinal.name(),GId[0],GId[1]))!='fail':
                    KEY = (Integer(GId[0]),Integer(GId[1]))
                    GStem = '%dgp%d'%KEY
                else:
                    raise ValueError
            except:
                if kwds.has_key('GroupDefinition'): 
                    KEY = (kwds['GroupDefinition'],)
                else:
                    if Hfinal.IsPermGroup():
                        KEY = ('Group('+repr(Hfinal.GeneratorsOfGroup())+')',)
                        KEY = (''.join([t.strip() for t in KEY[0].split()]),)
                    else:
                        if OPTION.opts['prot']:
                            print "computing regular permutation action"
                        KEY = (repr(GAP('regularPermutationAction(%s: forceDefiningGenerators)'%Hfinal.name())),)
                        KEY = (''.join([t.strip() for t in KEY[0].split()]),)
                        # (GAP.eval('Group(List([1..Length(GeneratorsOfGroup(%s))], x -> Image(IsomorphismPermGroup(%s),GeneratorsOfGroup(%s)[x])))'%(Hfinal.name(),Hfinal.name(),Hfinal.name())),)

        #########
        ## Do we have the non-commutative case? If Singular is less than 3-1-0, raise an error.
        if (q%2) or ((pr is not None) and (pr%2)):  
            SingVer = tuple([int(x) for x in singular.eval('system("version")')])
            if SingVer < (3,1,0):
                raise NotImplementedError, ("""There is only Singular %s installed.
We need at least Singular 3-1-0 in the non-commutative Case."""%('-'.join(len(SingVer)*['%d'])))%SingVer

        #########
        ## Determine the location of data storage
        root_explicit = kwds.get('root')
        if root_explicit is not None:
            if len(root_explicit):
                if root_explicit[-1]!='/':
                    root_explicit = root_explicit + '/'
            while '//' in root_explicit:
                root_explicit = root_explicit.replace('//','/')
        root_public_db = COHO.public_db 
        root_user_db = COHO.user_db #DOT_SAGE+'pGroupCohomology/db/'

        # KEY either provides the coordinates (q,n) of a group in the small
        # groups library, or is of the form (s,) with a string s such
        # that gap(s) defines a group with an appropriate generating set.
        # It can be hashed.
        # Moreover the stem name (GStem) is set up, and we may have 
        # a different GroupName (or None).

        # For non prime power groups, we need a sufficiently large subgroup.
        # In this case, Hfinal is there (even if KEY==(q,n))
        if pr is not None: # this is the case only for non-pp
            ## Check the input
            Subgroup = kwds.get('Subgroup')
            SubgpId = kwds.get('SubgpId')
            HP = kwds.get('SubgpCohomology')
            SylowSubgroup = kwds.get('SylowSubgroup')
            HSyl = kwds.get('SylowSubgpCohomology')
            
            # roughly check (Sylow)Subgroup
            if SylowSubgroup is not None:
                if not Hfinal.IsSubgroup(SylowSubgroup):
                    raise ValueError, "The given Sylow subgroup must be a subgroup"
                if pr.divides(Integer(gap.eval('Index(%s,%s)'%(Hfinal.name(),SylowSubgroup.name())))):
                    raise ValueError, "A Sylow %d-subgroup has index coprime to %d"%(pr,pr)
                if not pr.divides(Integer(gap.eval('Order(%s)'%SylowSubgroup.name()))):
                    raise ValueError, "The given Sylow subgroup's order is indivisible by %d"%pr
            if Subgroup is not None:
                if not Hfinal.IsSubgroup(Subgroup):
                    raise ValueError, "The given subgroup is in fact not a subgroup"
                if pr.divides(Integer(gap.eval('Index(%s,%s)'%(Hfinal.name(),Subgroup.name())))):
                    raise ValueError, "The given subgroup must contain a Sylow %d-subgroup"%pr
                if SylowSubgroup is not None:
                    if not Subgroup.IsSubgroup(SylowSubgroup):
                        raise ValueError, "The given subgroup must contain the given Sylow subgroup"

            # check HSyl and HP
            if HSyl is not None:
               if not isinstance(HSyl,COHO):
                   raise ValueError, "The given cohomology of a Sylow subgroup is not a cohomology ring"
               if isinstance(HSyl,MODCOHO):
                   raise ValueError, "The given cohomology of a Sylow subgroup does in fact not belong to a prime power group"
            if HP is not None: 
                if not isinstance(HP,COHO):
                    raise TypeError, "`SubgpCohomology` must be %s"%COHO
                HSyl = HP._HSyl or HP # ignore the keyword argument for HSyl

            ##################################
            # Begin to construct the basic data
            # First step: Get the (Sylow) subgroup related with the given cohomology
            phiSub = None
            phiSyl = None
            SubgroupTested = False
            SylowTested = False
            # 1a) Try to match with a given cohomology ring
            if Subgroup is None:
                if HP is not None:
                    try:
                        gap.RestoreStateRandom(0)
                        phiSub=gap('IsomorphicSubgroups(%s,%s:findall:=false)'%(HP.group().name(),Hfinal.name()))[1]
                        Subgroup = gap('Group(List([1..Length(GeneratorsOfGroup(%s))], x -> Image(%s, GeneratorsOfGroup(%s)[x])))'%(HP.group().name(),phiSub.name(),HP.group().name()))
                    except:
                        raise ValueError, "Unable to find a subgroup compatible with the argument `SubgpCohomology`"
                    SubgroupTested = True
            else:
                if HP is not None:
                    phiSub = HP.group().canonicalIsomorphism(Subgroup)
                    if repr(phiSub)=='fail':
                        raise ValueError, "The arguments `Subgroup` and `SubgpCohomology` don't match"
                    SubgroupTested=True
            # 1b) dito for the Sylow subgroup
            if SylowSubgroup is None:
                if (HP is not None) and (phiSub is not None):
                    SylowSubgroup = gap('Group(List([1..Length(GeneratorsOfGroup(%s))], x -> Image(%s, GeneratorsOfGroup(%s)[x])))'%((HP.sylow_subgroup or HP.group)().name(),phiSub.name(),(HP.sylow_subgroup or HP.group)().name()))
                    SylowTested = True
                elif HSyl is not None:
                    try:
                        SylowSubgroup = (Hfinal if Subgroup is None else Subgroup).SylowSubgroup(pr)
                        gap.RestoreStateRandom(0)
                        phiSyl = gap('IsomorphismGroups(%s,%s)'%(HSyl.group().name(),SylowSubgroup.name()))
                        SylowSubgroup = gap('Group(List([1..Length(GeneratorsOfGroup(%s))], x -> Image(%s, GeneratorsOfGroup(%s)[x])))'%(HSyl.group().name(),phiSyl.name(),HSyl.group().name()))
                    except:
                        raise ValueError, "Unable to find a Sylow subgroup compatible with the given arguments `SubgpCohomology` or `SylowSubgpCohomology`"
                    SylowTested = True
            else:
                if HSyl is not None:
                    phiSub = HSyl.group().canonicalIsomorphism(SylowSubgroup)
                    if repr(phiSub)=='fail':
                        raise ValueError, "The arguments `SylowSubgroup` and `SylowSubgpCohomology` don't match"
                    SylowTested=True


            #####
            # Second step: Get the cohomology from the given subgroups
            if SylowSubgroup is None:
                if OPTION.opts['prot']:
                    print "Try to compute a Sylow %d-subgroup"%pr
                SylowSubgroup = (Hfinal if Subgroup is None else Subgroup).SylowSubgroup(pr)
                # We are free in choosing generators, since apparently HSyl was not given
            if HSyl is None:
                try:
                    if OPTION.opts['prot']:
                        print "Try to find the SmallGroups address of the Sylow subgroup"
                    SylowId = SylowSubgroup.IdGroup().sage()
                except BaseException, msg:
                    if not ("group identification" in str(msg)):
                        raise msg
                    if OPTION.opts['prot']:
                        print "SmallGroups address not available. Compute the order"
                    SylowId = [Integer(SylowSubgroup.Order()),0]            
                if SylowId[1]>0:
                    gap.RestoreStateRandom(0)
                    phiSyl = gap('IsomorphismGroups(SmallGroup(%d,%d),%s)'%(SylowId[0],SylowId[1],SylowSubgroup.name()))
                    SylowSubgroup = gap('Group(List([1..Length(GeneratorsOfGroup(Source(%s)))],x->Image(%s,GeneratorsOfGroup(Source(%s))[x])))'%(phiSyl.name(),phiSyl.name(),phiSyl.name()))
                    HSyl = CohomologyRing(SylowId[0],SylowId[1],useElimination=kwds.get('useElimination'),auto=kwds.get('auto'),useFactorization=kwds.get('useFactorization'),root=kwds.get('root'))
                else:
                    if OPTION.opts['prot']:
                        print "Find a minimal generating set"
                    SylowSubgroup = SylowSubgroup.MinimalGeneratingSet().Group()                    
                    HSyl = CohomologyRing(SylowSubgroup,useElimination=kwds.get('useElimination'),auto=kwds.get('auto'),useFactorization=kwds.get('useFactorization'),root=kwds.get('root'), GroupName='SylowSubgroup(%s,%d)'%(GroupName or GStem,pr))
            # By now, we have HSyl and SylowSubgroup

            if kwds.get('OneStep'):
                Subgroup = SylowSubgroup
                HP = HSyl
                SubgpComputedFromScratch = False
            if Subgroup is None:
                if OPTION.opts['prot']:
                    print "Compute intermediate subgroup"
                Subgroup = Hfinal.Normalizer(SylowSubgroup.Centre())
                qP = Integer(Subgroup.Order())
                if qP==q or qP.is_prime_power():
                    # Subgroup=SylowSubgroup or =Hfinal
                    # In both cases, we are reduced to the OneStep case
                    Subgroup = SylowSubgroup
                    HP = HSyl
                    SubgpComputedFromScratch = False
                else:
                    SubgpComputedFromScratch = True
            else:
                SubgpComputedFromScratch = False

            if HP is None:
                try:
                    if OPTION.opts['prot']:
                        print "Try to find the SmallGroups address of the intermediate subgroup"
                    SubgpId = Subgroup.IdGroup().sage()
                except BaseException, msg:
                    if not ("group identification" in str(msg)):
                        raise msg
                    if OPTION.opts['prot']:
                        print "SmallGroups address not available. Compute the order"
                    SubgpId = [Integer(Subgroup.Order()),0]
                if SubgpId[1]>0: # SmallGroup name is better than my custom names
                    gap.RestoreStateRandom(0)
                    phiSub = gap('IsomorphismGroups(SmallGroup(%d,%d),%s)'%(SubgpId[0],SubgpId[1],Subgroup.name()))
                    Subgroup = gap('Group(List([1..Length(GeneratorsOfGroup(Source(%s)))],x->Image(%s,GeneratorsOfGroup(Source(%s))[x])))'%(phiSub.name(),phiSub.name(),phiSub.name()))
                    HP = CohomologyRing(Subgroup,SubgpId=SubgpId,prime=pr,SylowSubgroup=SylowSubgroup,SylowSubgpCohomology=HSyl,GStem='%dgp%d'%(SubgpId[0],SubgpId[1]), useElimination=kwds.get('useElimination'),useFactorization=kwds.get('useFactorization'))
                elif SubgpComputedFromScratch:
                    #print "SubgpComputedFromScratch"
                    # no minimal generating set needed
##                    if OPTION.opts['prot']:
##                        print "Find a minimal generating set"
##                    Subgroup = Subgroup.MinimalGeneratingSet().Group()
                    SubgpId=None
                    #print "SylowSubgroup",SylowSubgroup.GeneratorsOfGroup().Length()
                    #print "HSyl",HSyl.group().GeneratorsOfGroup().Length()
                    HP = CohomologyRing(Subgroup, prime=pr, SylowSubgpCohomology=HSyl, SylowSubgroup=SylowSubgroup, OneStep=True, GroupName='Normalizer(%s,Centre(SylowSubgroup(%s,%d)))'%(GroupName or GStem, GroupName or GStem,pr), useElimination=kwds.get('useElimination'),useFactorization=kwds.get('useFactorization'))
                else:
                    #print "else"
                    HP = CohomologyRing(Subgroup, prime=pr, SylowSubgpCohomology=HSyl, SylowSubgroup=SylowSubgroup, OneStep=True, GroupName='IntermediateSubgroup(%s,%d)'%(GroupName or GStem,pr), useElimination=kwds.get('useElimination'),useFactorization=kwds.get('useFactorization'))

            ############
            # By now, we have both subgroups and their cohomology rings.
            if not HP.completed:
                HP.make()
            # not needed for HSyl, since it was computed when we did it in the computation of HP

            ############
            # By now, SylowSubgroup is equal to HP.sylow_subgroup() under the canonical map from Subgroup to HP.group().
            # However, it is not necessarily true that their GENERATING SETS are related by the canonical map.
            # This will be taken care in MODCOHO.__init__.            

            ##################################
            # 
            # Extending the group key
            # if we have no GroupId, compute permutation representations
            #
            if len(KEY)==1:
                from pGroupCohomology.resolution import gap
                if not Hfinal.IsPermGroup():
                    GPerm = gap(KEY[0]) # gap('regularPermutationAction(%s: forceDefiningGenerators)'%Hfinal.name())
                    tmpPhi = gap('GroupHomomorphismByImages(%s,%s,GeneratorsOfGroup(%s),GeneratorsOfGroup(%s))'%(Hfinal.name(),GPerm.name(),Hfinal.name(),GPerm.name()))
                    PPerm = gap('Group(List([1..Length(GeneratorsOfGroup(%s))], x->Image(%s,GeneratorsOfGroup(%s)[x])))'%(Subgroup.name(),tmpPhi.name(),Subgroup.name()))
                else:
                    GPerm = Hfinal
                    PPerm = Subgroup
                    tmpPhi = None

        # We try to find the cohomology ring in the cache
        # or automatically reload it from disc.
        # We cover the non prime power case first:        
        OUT = None
        if pr is not None:
            CacheKey = (KEY, GStem, HP._key, pr)
            OUT = _cache.get(CacheKey)
            if OUT is None: ## Reload from disk. By default, we only save in user_db,
                            ## but an explicitely given root is fine as well.
                root = root_explicit or root_user_db
                if len(root) and root[-1]!='/':
                    root=root+'/'
                try:
                    OUT = load(root+'H'+GStem+'mod%d.sobj'%pr)
                except IOError:
                    pass
                if OUT is not None:
                    if not isinstance(OUT,MODCOHO):
                        raise TypeError, "We found %s on disk, not a modular cohomology ring"%type(OUT)
            if OUT is None:  # Unless the user forbids it, try to obtain it from some web source
                if kwds.get('websource')!=False:
                    try:
                        root = root_explicit or root_user_db
                        if isinstance(kwds.get('websource'), basestring):
                            OUT = self.web_db(GStem, websource=kwds.get('websource'), root=root)
                        else:
                            OUT = self.web_db(GStem, root=root)
                    except:
                        if OPTION.opts['prot']:
                            print "No cohomology ring found in repository. Starting from scratch."

            if OUT is not None:
                if OUT._key != CacheKey:
                    similarity = _IsKeyEquivalent(CacheKey,OUT._key)
                    if similarity: #(len(CacheKey[0])==1) and (repr(OUT.group().canonicalIsomorphism(OUT.group().parent()(CacheKey[0][0]))) != 'fail'):
                        if similarity == 1:
                            print 'Warning: Stored cohomology data have a different group description, but they seem to be equivalent'
                        return OUT
                    else:
                        raise ValueError, 'The Cohomology data stored at %s belong to a different group!'%(root+GStem+'.sobj')
                    
            if OUT is None: # need a computation from scratch
                if len(KEY)==1:
                    #print "KEY 1"
                    #print repr(HP)
                    #print Subgroup.GeneratorsOfGroup().Length()
                    #print HP.group().GeneratorsOfGroup().Length()
                    OUT = MODCOHO(Hfinal, pr, HP, Subgroup, GroupName=GroupName, GStem=GStem, GroupDescr=kwds.get('GroupDescr'), SubgpId=SubgpId, SubgpPerm=PPerm, GPerm=GPerm, useElimination=kwds.get('useElimination'),useFactorization=kwds.get('useFactorization'))
                else:
                    OUT = MODCOHO(Hfinal, pr, HP, Subgroup, GroupName=GroupName, GStem=GStem, GroupDescr=kwds.get('GroupDescr'), SubgpId=SubgpId, GroupId=KEY, useElimination=kwds.get('useElimination'),useFactorization=kwds.get('useFactorization'))
            if OUT._key != CacheKey:
                if len(OUT._key[0])==1:
                    GKEY = ''.join([t.strip() for t in OUT._key[0][0].split()])
                    tmpKey = list(OUT._key)
                    tmpKey[0] = (GKEY,)
                    OUT.setprop('_key',tuple(tmpKey))
                if OUT._key != CacheKey:
                    print OUT._key
                    print CacheKey
                    raise RuntimeError, "Keys for caching are corrupted"
                else:
                    from sage.all import save
                    if OPTION.opts['prot']:
                        print "Trying to update data on disk..."
                    save(OUT,OUT.autosave_name())
            #_cache[CacheKey] = OUT # not necessary, since MODCOHO.__init__ inserts into the cache
            return OUT

        # Remains the prime power case:
        if not kwds.get('websource'): # if a websource is provided, it will be used.
            if root_explicit is None:
                CacheKey = (KEY, root_public_db+GStem+'/dat/State')
                if _cache.has_key(CacheKey):
                    OUT = _cache[CacheKey]
                else:
                    CacheKey = (KEY, root_user_db+GStem+'/dat/State')
                    if _cache.has_key(CacheKey):
                        OUT = _cache[CacheKey]
            else:
                CacheKey = (KEY, root_explicit+GStem+'/dat/State')
                if _cache.has_key(CacheKey):
                    OUT = _cache[CacheKey]
            if OUT is not None:    # check if the singular subprocess still runs
                try:
                    foobar = OUT.GenS.parent().eval(OUT.GenS.name())
                except RuntimeError:
                    OUT = None
                    _cache.__delitem__(CacheKey)
            if OUT is not None:
                return OUT
            
            # It was not in the cache. So, we try to look it up on disk.
            if root_explicit is not None:
                try:
                    root = root_explicit
                    OUT = load(root_explicit+GStem+'/H'+GStem)
                except:
                    OUT = None
            else:
                try:
                    root = root_public_db
                    OUT = load(root_public_db+GStem+'/H'+GStem)
                except:
                    try:
                        root = root_user_db
                        OUT = load(root_user_db+GStem+'/H'+GStem)
                    except:
                        OUT = None

        if OUT is None:  # Unless the user forbids it, try to obtain it from some web source
            if kwds.get('websource')!=False:
                try:
                    root = root_explicit or root_user_db
                    if isinstance(kwds.get('websource'), basestring):
                        OUT = self.web_db(GStem, websource=kwds.get('websource'), root=root)
                    else:
                        OUT = self.web_db(GStem, root=root)
                except:
                    if OPTION.opts['prot']:
                        print "No cohomology ring found in repository. Starting from scratch."

        # Now put the result (if it was obtained yet!) into the cache
        if OUT is not None:
            StateData = root+'/'+GStem+'/dat/State'
            while '//' in StateData:
                StateData = StateData.replace('//','/')
            CacheKey = (KEY, StateData)
            
            if not isinstance(OUT, COHO):
                raise ValueError, 'The data stored at %s are no Cohomology data!'%(root+GStem+'/')
            if OUT.root != root:
                OUT.setprop('_need_new_root',1)
            if OUT._key != CacheKey:
                if (len(CacheKey[0])==1) and (repr(OUT.group().canonicalIsomorphism(OUT.group().parent()(CacheKey[0][0]))) != 'fail'):
                    print 'Warning: Stored cohomology data have a different group description, but they seem to be equivalent'
                    return OUT
                else:
                    raise ValueError, 'The Cohomology data stored at %s belong to a different group!'%(root+'/'+GStem+'/')
            if OUT is not _cache[CacheKey]:
                # Should not happen, since COHO_unpickle also does
                # caching, and even if data moved on disk then they
                # are inserted to the cache when being updated.
                # However, we try to get it right.
                OUT = _cache[CacheKey]
            return OUT
        

        #########################
        ## We can't help, the ring must be constructed from scratch.
        # Fancy optional parameters, such as 'auto', 'GroupDescr'
        # etc., will be used here, in ``extras``
        extras = {}
        for k in kwds.items():
            extras[k[0]] = k[1]
        extras['GroupName'] = GroupName
        extras['GStem'] = GStem
        #extras['key'] = CacheKey
        if len(KEY)==1:
            extras['gap_input'] = q # we must specify the group order
            if root_explicit is not None:
                extras['root'] = root_explicit
                CacheKey = (KEY, root_explicit+GStem+'/dat/State')
                extras['key'] = CacheKey
                extras['GStem'] = GStem
                OUT = COHO(gap(KEY[0]), **extras)
            else:
                try:
                    CacheKey = (KEY, root_public_db+GStem+'/dat/State')
                    extras['key'] = CacheKey
                    extras['root'] = root_public_db
                    extras['GStem'] = GStem
                    OUT = COHO(gap(KEY[0]), **extras)
                except:
                    if OPTION.opts['prot']:
                        print 'No write permission for the cohomology data base'
                        print '-- using private data base at', COHO.user_db
                    CacheKey = (KEY, root_user_db+GStem+'/dat/State')
                    extras['key'] = CacheKey
                    extras['root'] = root_user_db
                    extras['GStem'] = GStem
                    OUT = COHO(gap(KEY[0]), **extras)
        else:
            if root_explicit is not None:
                extras['root'] = root_explicit
                CacheKey = (KEY, root_explicit+GStem+'/dat/State')
                extras['key'] = CacheKey
                extras['GStem'] = GStem
                OUT = COHO(KEY[0],KEY[1], **extras)
            else:
                try:
                    CacheKey = (KEY, root_public_db+GStem+'/dat/State')
                    extras['key'] = CacheKey
                    extras['root'] = root_public_db
                    extras['GStem'] = GStem
                    OUT = COHO(KEY[0],KEY[1], **extras)
                except:
                    if OPTION.opts['prot']:
                        print 'No write permission for the cohomology data base'
                        print '-- using private data base at', COHO.user_db
                    CacheKey = (KEY, root_user_db+GStem+'/dat/State')
                    extras['key'] = CacheKey
                    extras['root'] = root_user_db
                    extras['GStem'] = GStem
                    OUT = COHO(KEY[0],KEY[1], **extras)
        if OUT._key != CacheKey:
            print CacheKey
            print OUT._key
            raise RuntimeError, "Keys for caching are corrupted"
        #_cache[CacheKey] = OUT # not necessary, since COHO.__init__ inserts into the cache
        return OUT

    def set_user_db(self, s = None):
        """
        Define the location of a user-dependent cohomology data base

        INPUT:

        ``s``, a string providing a folder name, or ``None``.

        If ``s`` is a string, a cohomology data base in the folder
        ``s`` will be activated as the user's private cohomology data
        base. If it is ``None``, a private data base in a default
        location will be activated.

        NOTE:

        If necessary, the folder will be created as soon as data from
        ``s`` are requested.

        EXAMPLES:

        We create a cohomology ring, whose data files are rooted in a
        temporary directory; it will be removed as soon as Sage is
        quit.
        ::

            sage: from pGroupCohomology import CohomologyRing
            sage: tmp_root = tmp_filename()
            sage: CohomologyRing.set_user_db(tmp_root)
            sage: H = CohomologyRing.user_db(8,3)

        Note that the given path ``tmp_root`` may contain double
        slashes.  :func:`~pGroupCohomology.CohomologyRing` replaces
        them by single slashes::

            sage: tmp_root = tmp_root.replace('//','/')
            sage: H.root.startswith(tmp_root)
            True

        """
        if s is None:
            s = DOT_SAGE+'/pGroupCohomology/db/'
        if not isinstance(s,basestring):
            raise TypeError, "String (pathname) expected"
        while '//' in s:
            s = s.replace('//','/')
        import os
        if os.path.exists(s):
            if not os.path.isdir(s):
                raise OSError, "There is a file %s that we won't overwrite"%s
        else:
            os.makedirs(s)
        COHO.user_db = s

    def user_db(self,*args, **kwds):
        """
        Retrieve a cohomology ring from a user-dependent cohomology data base

        EXAMPLES:

        We create a cohomology ring, whose data files are rooted in a
        temporary directory; it will be removed as soon as Sage is
        quit. We use the optional parameter ``websource=False`` in
        order to ensure that it is not tried to download a completely
        computed cohomology ring from the web.
        ::

            sage: from pGroupCohomology import CohomologyRing
            sage: tmp_root = tmp_filename()
            sage: CohomologyRing.set_user_db(tmp_root)
            sage: H = CohomologyRing.user_db(8,3, websource=False)
            sage: print H
            Cohomology ring of Dihedral group of order 8 with coefficients in GF(2)
            <BLANKLINE>
            Computed up to degree 0
            Minimal list of generators:
            []
            Minimal list of algebraic relations:
            []

        """
        kwds['root'] = COHO.user_db
        return self(*args, **kwds)

    def set_web_db(self, s = None):
        """
        Redefine the default location of a web source for cohomology rings

        INPUT:

        ``s``, a string providing a folder name, or ``None``.

        If ``s`` is a string, then cohomology rings will be searched
        in a web source named by ``s``. If it is ``None``, the web
        source is reset to some default location.
        
        EXAMPLES:

        The example produces files. For safety reasons, we choose
        files in a temporary directory; it will be removed as soon as
        Sage is quit.
        ::

            sage: tmp_root = tmp_filename()
            sage: from pGroupCohomology import CohomologyRing
            sage: CohomologyRing.set_web_db('http://sage.math.washington.edu/home/SimonKing/Cohomology/db/')
            sage: H=CohomologyRing.web_db('8gp3', root=tmp_root)
            sage: print H
            Cohomology ring of Dihedral group of order 8 with coefficients in GF(2)
            <BLANKLINE>
            Computation complete
            Minimal list of generators:
            [c_2_2: 2-Cocycle in H^*(D8; GF(2)),
             b_1_0: 1-Cocycle in H^*(D8; GF(2)),
             b_1_1: 1-Cocycle in H^*(D8; GF(2))]
            Minimal list of algebraic relations:
            [b_1_0*b_1_1]
        
        """
        from pGroupCohomology.cohomology import COHO
        if s is None:
            s = 'http://sage.math.washington.edu/home/SimonKing/Cohomology/db/'
        if isinstance(s,basestring):
            if not s.endswith('/'):
                s = s+'/'
            COHO.web_db = s
        else:
            raise ValueError, "String expected"    


    # TODO: non prime power groups
    def web_db(self, GStem, websource = None, prime=None, root = None):
        """
        Import a cohomology ring from a web source.

        NOTE:

        Usually this function would not be directly used. It is
        automatically called by
        :func:`~pGroupCohomology.CohomologyRing` if a cohomology ring
        can not be found in a local folder.

        INPUT:

        - ``GStem``, a string so that ``GStem+'.tar.gz'`` can be found
          in the web source, if it is a prime power group, or
          ``'H'+GStem+'mod%d.sobj'%prime`` otherwise.
        - ``websource``: If ``None`` (default), a default location
          (respectively the location provided by
          :meth:`~pGroupCohomology.CohomologyRingFactory.set_web_db`)
          is chosen. Otherwise, it should be a web address.
        - ``prime``: An optional prime, the modulus of the cohomology
          ring. It must be provided if ond *only* if the group is not
          a prime power group.
        - ``root``: If ``None`` (default), the data are unpacked into
          the user's currently activated private cohomology data base.
          Otherwise, it provides a directory into which the downloaded
          cohomology data will be unpacked.

        EXAMPLES:

        The example produces files. For safety reasons, we choose
        files in a temporary directory; it will be removed as soon as
        Sage is quit. We use a protocol mode, so that it is visible
        what happens behind the scenes.
        ::

            sage: tmp_root = tmp_filename()
            sage: from pGroupCohomology import CohomologyRing
            sage: from pGroupCohomology.resolution import OPTION
            sage: OPTION('prot')
            <BLANKLINE>
            sage: H=CohomologyRing.web_db('8gp3', websource='http://sage.math.washington.edu/home/SimonKing/Cohomology/db/', root=tmp_root)
            Downloading http://sage.math.washington.edu/home/SimonKing/Cohomology/db/8gp3.tar.gz
            Extracting archive file at ...
            The state of this cohomology ring is expected to be provided at .../8gp3/dat/State.sobj

            sage: print H
            Cohomology ring of Dihedral group of order 8 with coefficients in GF(2)
            <BLANKLINE>
            Computation complete
            Minimal list of generators:
            [c_2_2: 2-Cocycle in H^*(D8; GF(2)),
             b_1_0: 1-Cocycle in H^*(D8; GF(2)),
             b_1_1: 1-Cocycle in H^*(D8; GF(2))]
            Minimal list of algebraic relations:
            [b_1_0*b_1_1]

        In order to not break the other doc tests, we now switch the
        protocol mode off.
        ::
        
            sage: H.option('noprot')
              
        """
##        raise RuntimeError
        if root is None:
            root = COHO.user_db
        if not root.endswith('/'):
            root = root + '/'
        while '//' in root:
            root = root.replace('//','/')
            
        if websource is None:
            websource = COHO.web_db
        if not websource.endswith('/'):
            websource = websource + '/'
        
        # First step: Download the tar file from the web and unpack it to root
        import urllib
        import tarfile
        from pGroupCohomology.resolution import OPTION
        if prime is None:
            if OPTION.opts['prot']:
                print "Downloading "+websource + GStem + '.tar.gz'
            (f, h) = urllib.urlretrieve(websource + GStem + '.tar.gz')
            if OPTION.opts['prot']:
                print "Extracting archive file at "+f
            T = tarfile.open(f)
            T.extractall(path=root)
        else:
            if not (hasattr(prime,'is_prime') and prime.is_prime()):
                raise ValueError, '``prime`` must be a prime number'
            if OPTION.opts['prot']:
                print "Downloading "+websource + 'H'+GStem + 'mod%d.sobj'%prime
            (f, h) = urllib.urlretrieve(websource + 'H'+GStem + 'mod%d.sobj'%prime)
            try:
                OUT = load(f)
                if isinstance(OUT,COHO):
                    GStemList = GStem.split('gp')
                    if len(GStemList)==2:
                        if GStemList[0].isdigit() and GStemList[1].isdigit():
                            q = int(GStemList[0])
                            n = int(GStemList[1])
                            if OUT.GroupNames.has_key((q,n)):
                                if OUT.GroupName!=OUT.GroupNames[q,n][0] or OUT.GroupDescr!=OUT.GroupNames[q,n][1]:
                                    OUT.setprop('GroupName',OUT.GroupNames[q,n][0])
                                    OUT.setprop('GroupDescr',OUT.GroupNames[q,n][1])
                    save(OUT,root+'H'+GStem + 'mod%d.sobj'%prime)
                    return OUT
                raise ValueError
            except:
                raise RuntimeError, "Unable to import cohomology ring from "+websource + 'H'+GStem + 'mod%d.sobj'%prime            

        ## Second step: load the cohomology ring and return it
        ## It is now the prime power case.
        OPTION.opts['@use_this_root@'] = root
        try:
            OUT = load(root + GStem + '/H'+GStem)
            r = OUT.root # this line may imply that the downloaded data are changed on disk in order to make them fit
        except:
            raise RuntimeError, "Unable to import cohomology ring from "+websource + GStem + '.tar.gz'
        if isinstance(OUT,COHO):
            GStemList = GStem.split('gp')
            if len(GStemList)==2:
                if GStemList[0].isdigit() and GStemList[1].isdigit():
                    q = int(GStemList[0])
                    n = int(GStemList[1])
                    if OUT.GroupNames.has_key((q,n)):
                        if OUT.GroupName!=OUT.GroupNames[q,n][0] or OUT.GroupDescr!=OUT.GroupNames[q,n][1]:
                            OUT.setprop('GroupName',OUT.GroupNames[q,n][0])
                            OUT.setprop('GroupDescr',OUT.GroupNames[q,n][1])
                            from sage.all import save
                            save(OUT, OUT.autosave_name())
        else:
            raise RuntimeError, "Unable to import cohomology ring from "+websource + GStem + '.tar.gz'
        return OUT


def _IsKeyEquivalent(k1, k2):
    """
    Test equivalence of keys of cohomology rings

    INPUT:

    ``k1``, ``k1``: Keys of cohomology rings

    OUTPUT:
    
    - 0, if the keys are essentially different,
    - 1 if they are equivalent,
    - 2 if they are equal (up to file location)

    NOTE:

    If the keys are equivalent, the ring presentations of the cohomology ring
    should be equal.

    EXAMPLES::

        sage: from pGroupCohomology import CohomologyRing, _IsKeyEquivalent
        sage: tmp_root = tmp_filename()
        sage: CohomologyRing.set_user_db(tmp_root)
        sage: G = gap('SymmetricGroup(6)')
        sage: G.IdGroup()
        [ 720, 763 ]
        sage: H1 = CohomologyRing(G,prime=2,GroupName='Sym6')
        sage: H2 = CohomologyRing(720,763,prime=2)
        sage: _IsKeyEquivalent(H1._key,H2._key)
        0

    In fact, mapping the generators of ``H1.group()`` to the generators
    of ``H2.group()`` does not result in a group isomorphism. This is tested as
    follows::

        sage: H1.group().canonicalIsomorphism(H2.group())
        fail

    So, we chose a different generating system for ``G``. In order to
    have a reproducible doc test, we choose an explicit group isomorphism::

        sage: phiG = gap('GroupHomomorphismByImages( Group([(1,2),(1,2,3,4,5,6)]), SymmetricGroup([ 1 .. 6 ]), [(1,2),(2,3,5,6,4)], [(5,6),(1,2,3,4,5)])')
        sage: Gnew = gap('Group(List([1..Length(GeneratorsOfGroup(%s))],x->Image(%s,GeneratorsOfGroup(%s)[x])))'%(H2.group().name(),phiG.name(),H2.group().name()))
        sage: H1new = CohomologyRing(Gnew,prime=2,GroupName='Sym6New')

    Now, the group keys are in fact essentially equal, since the
    program realises that the generators of Gnew correspond to those
    of ``SmallGroup(720,763)`` and thus identifies both rings::

        sage: _IsKeyEquivalent(H1new._key,H2._key)
        2
        sage: H2 is H1new
        True

    Now, we do the opposite and chose a generating set for
    ``SmallGroup(720,763)`` equivalent to the generating set for
    ``G``. Again, we define the isomorphism explicitly::

        sage: G2 = gap('SmallGroup(720,763)')
        sage: phiG2 = gap('GroupHomomorphismByImages( SymmetricGroup([ 1 .. 6 ]), Group([(1,2),(1,2,3,4,5,6)]), [(5,6),(1,2,3,4,5)], [(1,2),(2,6,3,4,5)])')
        sage: G2new = gap('Group(List([1..Length(GeneratorsOfGroup(%s))],x->Image(%s,GeneratorsOfGroup(%s)[x])))'%(G.name(),phiG2.name(),G.name()))
        sage: H2new = CohomologyRing(G2new,prime=2,GroupName='Sym6New2')

    Now the keys of the cohomology ring are equivalent, but not equal::

        sage: _IsKeyEquivalent(H2new._key,H1._key)
        1
        sage: H2new._key
        (('Group([(1,6,3,4,5,2),(3,6)])',), 'Sym6New2', ((16, 11), '.../16gp11/dat/State'), 2)
        sage: H1._key
        (('Group([(1,2,3,4,5,6),(1,2)])',), 'Sym6', ((16, 11), '.../16gp11/dat/State'), 2)

    """
    from pGroupCohomology.resolution import gap
    if len(k1)!=len(k2):
        return 0
    if k1[0]==k2[0]:
        similarity = 2
    else:
        if len(k1[0])==1:
            G1 = gap(k1[0][0])
        else:
            G1 = gap('SmallGroup(%d,%d)'%(k1[0][0],k1[0][1]))
        if len(k2[0])==1:
            G2 = gap(k2[0][0])
        else:
            G2 = gap('SmallGroup(%d,%d)'%(k2[0][0],k2[0][1]))
        if repr(G1.canonicalIsomorphism(G2))=='fail':
            return 0
        else:
            similarity = 1
    if len(k1)==3:
        return min(similarity, _IsKeyEquivalent(k1[-1],k2[-1]))
    return similarity

CohomologyRing = CohomologyRingFactory()
CohomologyRing.__doc__ = r"""
Constructor for modular cohomology rings of finite p-groups

This constructor is an instance of :class:`~pGroupCohomology.CohomologyRingFactory`.
See there for more examples.

The constructor can be called directly. Then, it is first checked
whether the completely computed cohomology ring of the given group is
part of some data base, or whether it can be downloaded. If this is
not the case, a new cohomology ring is created, being part of the
public or a user defined data base.

Using :meth:`~pGroupCohomology.CohomologyRingFactory.set_user_db`, the
location of the user defined data base can be determined. By
:meth:`~pGroupCohomology.CohomologyRingFactory.user_db`, one can
explicitly ask for taking data from the user defined data base. The
input formats for calling :func:`~pGroupCohomology.CohomologyRing` and
for calling :meth:`~pGroupCohomology.CohomologyRingFactory.user_db`
are the same.

INPUT:

**Parameters describing the group**

- a finite `p`-group `G` , either

  * given by a prime power ``q`` and a positive number ``n``, determining
    an entry of the SmallGroups library, or
  * given as an object in the Gap interface  
- ``GroupName`` (optional string): a name for the group. If the 
  group `G` is given in the Gap interface and if it is not provided with
  a custom name (using Gap's ``SetName``) then ``GroupName`` *must* be
  provided.
- ``GroupDescr`` (optional string): a description of the group. This can be
  any string, and is used when printing the cohomology ring or creating a
  web-site for it.

**Parameters describing the data base**

- ``root``: A string that, when given, provides a base folder for the data
  files. This folder is created, if necessary.
- ``websource``: If it is ``False``, it is not attempted to download data
  from some data base in the web. If it is a string providing the location
  of a data base in the web, then it is attempted to download the data from
  there. If ``websource`` is not given then first it is tried to look up
  data in the local file system, and if this fails then it is attempted to
  download the data from some default location in the web.

**Parameters modifying the algorithm**

- ``useElimination`` (optional, default ``None``): If ``True``, the 
  elimination method is used for trying to construct the Dickson classes.
  If ``False``, the linear algebra method is used for that purpose. By
  default, the linear algebra method is chosen, unless there is a Dickson
  class in degree greater than 18 (for prime power groups) or 20 (for non
  prime power groups).
- ``DicksonExp`` (optional integer, default = 3): If the elimination
  method for finding the Dickson classes is used, it is needed to set a
  bound for the power to which the Dickson classes might be raised: If
  `G` is a `p`-group and `n` is the given ``DicksonExp``, then the
  Dickson classes of the elementary abelian sub-groups of `G` are raised
  to the power of `p^0,p^1,...,p^n` before trying to simultaneously lift 
  them to `G`. We do not know any example in which the default value would
  not suffice.
- ``useFactorization`` (optional boolean, default True): Try to replace
  the Dickson classes of `G` by their minimal non-constant factors. This
  may simplify some computations, but there are rare examples in which the
  factorisation is a bigger problem than a high degree bound.
- ``auto`` (optional integer, default = 2 for abelian groups, and = 4
  otherwise): Only applies to the case of prime power groups. A quick but
  potentially memory consuming method for lifting chain maps will be 
  used in degree at most ``auto``. For prime power groups up to orders
  around 500, the default value seems to be heuristically best.
- ``useSlimgb`` / ``useStd`` (optinal boolean): Use Singular's ``slimgb``
  (resp. ``std``) for computing the Groebner basis of the relation ideal.
  If not specified, Singular's ``groebner`` method is chosen, which uses
  a heuristics to find the best algorithm for the computation of the
  Groebner basis.

**Global options**

``options`` (optional string or list of strings): set/unset global options.

There are various options, e.g. concerning protocol output, that apply to *all*
cohomology rings when set for one ring. Each option is set by a string, and unset
by prepending ``'no'`` to that string.

  **Available options**

  * ``'prot'`` [not default], display protocol
  * ``'timing'`` [not default], display timing
  * ``'useMTX'`` [default], use :class:`~pGroupCohomology.mtx.MTX`
    matrices for linear algebra over finite fields, when computing
    the ring structure. Note that the resolutions will always be
    computed using the Aachen
    `C MeatAxe <http://www.math.rwth-aachen.de/homes/MTX/>`_. By
    consequence, if ``useMTX`` is turned off, much time is spent
    on the conversion of different matrix types.
  * ``'save'`` [default], save intermediate results
  * ``'sparse'`` [default], remove temporarily unneeded data from
    memory 
  * ``'liftlist'`` [not default], try to lift a whole bunch of
    cochains at once

The options can also be (un)set later, by using the method 
:meth:`~pGroupCohomology.cohomology.COHO.option`.

The options ``'prot'`` and ``'timing'`` are useful to see the progress
of lengthy computations. If ``'save'`` is set, the ring will be saved
after each computed degree. This will come in handy if a long computation
was interrupted. Note, however, that many data, including a minimal
projective resolution for prime power groups, will *always* be saved on
disk.

The following options only apply to prime power groups, are somehow
experimental, and are not officially supported.

* If ``'useMTX'`` is unset, the computation time usually dramatically
  increases. So, we strongly recommend to not unset ``'useMTX'``.
* If ``'save'`` is set, the ring will be saved after each computed degree.
  This will come in handy if a long computation was interrupted. Note,
  however, that many data, including a minimal projective resolution, will
  *always* be saved on disk.
* Using the option ``'sparse'`` not only allows for computation of rather
  big examples; it turns out that it improves the computation time in small 
  examples. So, it is strongly recommended to not unset ``'sparse'``.
* With the option ``'liftlist'``, some examples are faster, others are slower. 
  On the groups of order 64, it is slightly faster to not use this option, 
  therefore our default is to not use it.

"""
